// ---------------------------------------------------------------------
//
// JavaBridge
//
// Copyright (c) 2012-2021 WINSOFT
//
// ---------------------------------------------------------------------

unit ClassFile;

{$ifdef CONDITIONALEXPRESSIONS}
  {$if CompilerVersion >= 14}
    {$define D6PLUS} // Delphi 6 or higher
  {$ifend}

  {$if CompilerVersion >= 17}
    {$define D2005PLUS} // Delphi 2005 or higher
  {$ifend}

  {$if CompilerVersion >= 20}
    {$define D2009PLUS} // Delphi 2009 or higher
  {$ifend}

  {$if CompilerVersion >= 21}
    {$define D2010PLUS} // Delphi 2010 or higher
  {$ifend}

  {$if CompilerVersion >= 24}
    {$define DXE3PLUS} // Delphi XE3 or higher
  {$ifend}
{$endif CONDITIONALEXPRESSIONS}

interface

uses ClassFileConstants;

type
{$ifndef D6PLUS}
  PWord = ^Word;
  PLongWord = ^LongWord;
  PSingle = ^Single;
  PDouble = ^Double;
{$endif D6PLUS}

{$ifdef D2009PLUS}
  TString = string;
{$else}
  TString = WideString;
{$endif D2009PLUS}

  UInt16 = Word;
  UInt32 = LongWord;
{$ifdef D2010PLUS}
  TByteDynArray = TArray<Byte>;
  TWordDynArray = TArray<Word>;
  TIntegerDynArray = TArray<Integer>;
{$else}
  UInt64 = Int64;
  PUInt64 = ^UInt64;
  TByteDynArray = array of Byte;
  TWordDynArray = array of Word;
  TIntegerDynArray = array of Integer;
{$endif D2010PLUS}

  // Signature

  TBaseType = (btUnknown, btByte, btChar, btDouble, btFloat, btInt, btLong, btShort, btBoolean);

  TClassFile = class;

  TClassFileItem = class {$ifdef D2005PLUS} abstract {$endif D2005PLUS}
  private
    FClassFile: TClassFile;
  protected
    constructor Create(ClassFile: TClassFile);
  public
    property ClassFile: TClassFile read FClassFile;
  end;

  TSignature = class {$ifdef D2005PLUS} abstract {$endif D2005PLUS} (TClassFileItem)
  protected
    class function CreateSignature(ClassFile: TClassFile; const Signature: TString; var Index: Integer): TSignature; overload;
    class function CreateSignature(ClassFile: TClassFile; const Signature: TString): TSignature; overload;
    constructor Create(ClassFile: TClassFile);
  end;

  TSignatureDynArray = array of TSignature;

  TBaseSignature = class(TSignature)
  private
    FBaseType: TBaseType;
  protected
    constructor Create(ClassFile: TClassFile; BaseType: TBaseType);
  public
    property BaseType: TBaseType read FBaseType;
  end;

  TObjectSignature = class(TSignature)
  private
    FName: TString;
  protected
    constructor Create(ClassFile: TClassFile; const Name: TString);
  public
    property Name: TString read FName;
  end;

  TArraySignature = class(TSignature)
  private
    FDimension: Integer;
    FItemSignature: TSignature;
  protected
    constructor Create(ClassFile: TClassFile; Dimension: Integer; ItemSignature: TSignature);
  public
    destructor Destroy; override;
    property Dimension: Integer read FDimension;
    property ItemSignature: TSignature read FItemSignature;
  end;

  TMethodSignature = class(TSignature)
  private
    FParameters: TSignatureDynArray;
    FResultType: TSignature;
    function GetParameterCount: Integer;
    function GetParameter(Index: Integer): TSignature;
  protected
    constructor Create(ClassFile: TClassFile; Parameters: TSignatureDynArray; ResultType: TSignature);
  public
    destructor Destroy; override;
    property ParameterCount: Integer read GetParameterCount;
    property Parameter[Index: Integer]: TSignature read GetParameter;
    property ResultType: TSignature read FResultType;
  end;

  // Constant Pool

  TConstantPoolItem = class {$ifdef D2005PLUS} abstract {$endif D2005PLUS} (TClassFileItem)
  protected
    constructor Create(ClassFile: TClassFile);
  end;

  TConstantClass = class(TConstantPoolItem)
  private
    FNameIndex: Word;
    function GetName: TString;
  protected
    constructor Create(ClassFile: TClassFile; var Offset: Integer);
  public
    property NameIndex: Word read FNameIndex;
    property Name: TString read GetName;
  end;

  TConstantInteger = class(TConstantPoolItem)
  private
    FValue: LongWord;
  protected
    constructor Create(ClassFile: TClassFile; var Offset: Integer);
  public
    property Value: LongWord read FValue;
  end;

  TConstantFloat = class(TConstantPoolItem)
  private
    FValue: Single;
  protected
    constructor Create(ClassFile: TClassFile; var Offset: Integer);
  public
    property Value: Single read FValue;
  end;

  TConstantLong = class(TConstantPoolItem)
  private
    FValue: UInt64;
  protected
    constructor Create(ClassFile: TClassFile; var Offset: Integer);
  public
    property Value: UInt64 read FValue;
  end;

  TConstantDouble = class(TConstantPoolItem)
  private
    FValue: Double;
  protected
    constructor Create(ClassFile: TClassFile; var Offset: Integer);
  public
    property Value: Double read FValue;
  end;

  TConstantFieldRef = class(TConstantPoolItem)
  private
    FClassIndex: Word;
    FNameAndTypeIndex: Word;
    function GetClassValue: TString;
    function GetName: TString;
    function GetTypeName: TString;
    function GetTypeSignature: TSignature;
  protected
    constructor Create(ClassFile: TClassFile; var Offset: Integer);
  public
    property ClassIndex: Word read FClassIndex;
    property NameAndTypeIndex: Word read FNameAndTypeIndex;
    property ClassValue: TString read GetClassValue;
    property Name: TString read GetName;
    property TypeName: TString read GetTypeName;
    property TypeSignature: TSignature read GetTypeSignature;
  end;

  TConstantMethodRef = class(TConstantPoolItem)
  private
    FClassIndex: Word;
    FNameAndTypeIndex: Word;
    function GetClassValue: TString;
    function GetName: TString;
    function GetTypeName: TString;
    function GetTypeSignature: TSignature;
  protected
    constructor Create(ClassFile: TClassFile; var Offset: Integer);
  public
    property ClassIndex: Word read FClassIndex;
    property NameAndTypeIndex: Word read FNameAndTypeIndex;
    property ClassValue: TString read GetClassValue;
    property Name: TString read GetName;
    property TypeName: TString read GetTypeName;
    property TypeSignature: TSignature read GetTypeSignature;
  end;

  TConstantInterfaceMethodRef = class(TConstantPoolItem)
  private
    FClassIndex: Word;
    FNameAndTypeIndex: Word;
  protected
    constructor Create(ClassFile: TClassFile; var Offset: Integer);
  public
    property ClassIndex: Word read FClassIndex;
    property NameAndTypeIndex: Word read FNameAndTypeIndex;
  end;

  TConstantString = class(TConstantPoolItem)
  private
    FStringIndex: Word;
    function GetStringValue: TString;
  protected
    constructor Create(ClassFile: TClassFile; var Offset: Integer);
  public
    property StringIndex: Word read FStringIndex;
    property StringValue: TString read GetStringValue;
  end;

  TConstantUtf8 = class(TConstantPoolItem)
  private
    FValue: TString;
    FSignature: TSignature;
    function GetSignature: TSignature;
    function GetMethodSignature: TMethodSignature;
    procedure CheckByte(Value: Byte);
  protected
    constructor Create(ClassFile: TClassFile; var Offset: Integer);
  public
    property Value: TString read FValue;
    property Signature: TSignature read GetSignature;
    property MethodSignature: TMethodSignature read GetMethodSignature;
  end;

  TConstantNameAndType = class(TConstantPoolItem)
  private
    FNameIndex: Word;
    FDescriptorIndex: Word;
    function GetDescriptor: TString;
    function GetName: TString;
    function GetSignature: TSignature;
  protected
    constructor Create(ClassFile: TClassFile; var Offset: Integer);
  public
    property NameIndex: Word read FNameIndex;
    property DescriptorIndex: Word read FDescriptorIndex;
    property Name: TString read GetName;
    property Descriptor: TString read GetDescriptor;
    property Signature: TSignature read GetSignature;
  end;

  TConstantMethodHandle = class(TConstantPoolItem)
  private
    FReferenceKind: Byte;
    FReferenceIndex: Word;
  protected
    constructor Create(ClassFile: TClassFile; var Offset: Integer);
  public
    property ReferenceKind: Byte read FReferenceKind;
    property ReferenceIndex: Word read FReferenceIndex;
  end;

  TConstantMethodType = class(TConstantPoolItem)
  private
    FDescriptorIndex: Word;
    function GetDescriptor: TString;
  protected
    constructor Create(ClassFile: TClassFile; var Offset: Integer);
  public
    property DescriptorIndex: Word read FDescriptorIndex;
    property Descriptor: TString read GetDescriptor;
  end;

  TConstantInvokeDynamic = class(TConstantPoolItem)
  private
    FBootstrapMethodAttrIndex: Word;
    FNameAndTypeIndex: Word;
  protected
    constructor Create(ClassFile: TClassFile; var Offset: Integer);
  public
    property BootstrapMethodAttrIndex: Word read FBootstrapMethodAttrIndex;
    property NameAndTypeIndex: Word read FNameAndTypeIndex;
  end;

  TConstantPoolItemDynArray = array of TConstantPoolItem;

  // Attribute

  TAttribute = class(TClassFileItem)
  private
    FLength: LongWord;
    FNameIndex: Word;
    FStart: Integer;
    function GetName: TString;
  protected
    class function CreateAttribute(ClassFile: TClassFile; var Offset: Integer): TAttribute;
    constructor Create(ClassFile: TClassFile; Start: Integer);
  public
    property NameIndex: Word read FNameIndex;
    property Name: TString read GetName;
    property Start: Integer read FStart;
    property Length: LongWord read FLength;
  end;

  TAttributeDynArray = array of TAttribute;

  TConstantValueAttribute = class(TAttribute)
  private
    FValueIndex: Word;
    function GetValue: TConstantPoolItem;
  protected
    constructor Create(ClassFile: TClassFile; var Offset: Integer);
  public
    property ValueIndex: Word read FValueIndex;
    property Value: TConstantPoolItem read GetValue;
  end;

  TSourceFileAttribute = class(TAttribute)
  private
    FSourceFileIndex: Word;
    function GetSourceFile: TString;
  protected
    constructor Create(ClassFile: TClassFile; var Offset: Integer);
  public
    property SourceFileIndex: Word read FSourceFileIndex;
    property SourceFile: TString read GetSourceFile;
  end;

  TException = record
    StartPc: Word;
    EndPc: Word;
    HandlerPc: Word;
    CatchType: Word;
  end;

  TExceptionDynArray = array of TException;

  TCodeAttribute = class(TAttribute)
  private
    FMaxStack: Word;
    FMaxLocals: Word;
    FCodeLength: LongWord;
    FExceptions: TExceptionDynArray;
    FAttributes: TAttributeDynArray;
    function GetCode(Index: Integer): Byte;
    function GetAttribute(Index: Integer): TAttribute;
    function GetAttributeCount: Integer;
    function GetException(Index: Integer): TException;
    function GetExceptionCount: Integer;
  protected
    constructor Create(ClassFile: TClassFile; var Offset: Integer);
  public
    property MaxStack: Word read FMaxStack;
    property MaxLocals: Word read FMaxLocals;
    property CodeLength: LongWord read FCodeLength;
    property Code[Index: Integer]: Byte read GetCode; default;
    property ExceptionCount: Integer read GetExceptionCount;
    property Exception[Index: Integer]: TException read GetException;
    property AttributeCount: Integer read GetAttributeCount;
    property Attribute[Index: Integer]: TAttribute read GetAttribute;
  end;

  TDeprecatedAttribute = class(TAttribute)
  protected
    constructor Create(ClassFile: TClassFile; var Offset: Integer);
  end;

  TSignatureAttribute = class(TAttribute)
  private
    FDescriptorIndex: Word;
    function GetDescriptor: TConstantUtf8;
  protected
    constructor Create(ClassFile: TClassFile; var Offset: Integer);
  public
    property DescriptorIndex: Word read FDescriptorIndex;
    property Descriptor: TConstantUtf8 read GetDescriptor;
  end;

  TInnerClass = record
    InnerClassIndex: Word;
    OuterClassIndex: Word;
    InnerNameIndex: Word;
    InnerClassAccessFlags: Word;
  end;

  TInnerClassDynArray = array of TInnerClass;

  TInnerClassesAttribute = class(TAttribute)
  private
    FClasses: TInnerClassDynArray;
    function GetInnerClass(Index: Integer): TInnerClass;
    function GetInnerClassCount: Integer;
    function GetInnerName(Index: Integer): TString;
  protected
    constructor Create(ClassFile: TClassFile; var Offset: Integer);
  public
    property Classes: TInnerClassDynArray read FClasses;
    property InnerClassCount: Integer read GetInnerClassCount;
    property InnerClass[Index: Integer]: TInnerClass read GetInnerClass;
    property InnerName[Index: Integer]: TString read GetInnerName;
  end;

  TLineNumber = record
    StartPc: Word;
    LineNumber: Word;
  end;

  TLineNumberDynArray = array of TLineNumber;

  TLineNumberTableAttribute = class(TAttribute)
  private
    FLineNumbers: TLineNumberDynArray;
    function GetLineNumber(Index: Integer): TLineNumber;
    function GetLineNumberCount: Integer;
  protected
    constructor Create(ClassFile: TClassFile; var Offset: Integer);
  public
    property LineNumberCount: Integer read GetLineNumberCount;
    property LineNumber[Index: Integer]: TLineNumber read GetLineNumber; default;
  end;

  TLocalVariable = record
    StartPc: Word;
    Length: Word;
    NameIndex: Word;
    DescriptorIndex: Word;
    Index: Word;
  end;

  TLocalVariableDynArray = array of TLocalVariable;

  TLocalVariableTableAttribute = class(TAttribute)
  private
    FLocalVariables: TLocalVariableDynArray;
    function GetLocalVariableCount: Integer;
    function GetLocalVariable(Index: Integer): TLocalVariable;
    function GetLocalVariableName(Index: Integer): TString;
    function GetLocalVariableType(Index: Integer): TString;
    procedure CheckIndex(Index: Integer);
  protected
    constructor Create(ClassFile: TClassFile; var Offset: Integer);
  public
    property LocalVariableCount: Integer read GetLocalVariableCount;
    property LocalVariable[Index: Integer]: TLocalVariable read GetLocalVariable; default;
    property LocalVariableName[Index: Integer]: TString read GetLocalVariableName;
    property LocalVariableType[Index: Integer]: TString read GetLocalVariableType;
  end;

  TMethodParameter = record
    NameIndex: Word;
    AccessFlags: Word;
  end;

  TMethodParameterDynArray = array of TMethodParameter;

  TMethodParametersAttribute = class(TAttribute)
  private
    FMethodParameters: TMethodParameterDynArray;
    procedure CheckIndex(Index: Integer);
    function GetMethodParameterCount: Integer;
    function GetMethodParameter(Index: Integer): TMethodParameter;
    function GetMethodParameterName(Index: Integer): TString;
    function GetIsMethodParameterFinal(Index: Integer): Boolean;
    function GetIsMethodParameterSynthetic(Index: Integer): Boolean;
    function GetIsMethodParameterMandated(Index: Integer): Boolean;
  protected
    constructor Create(ClassFile: TClassFile; var Offset: Integer);
  public
    property MethodParameterCount: Integer read GetMethodParameterCount;
    property MethodParameter[Index: Integer]: TMethodParameter read GetMethodParameter; default;
    property MethodParameterName[Index: Integer]: TString read GetMethodParameterName;
    property IsMethodParameterFinal[Index: Integer]: Boolean read GetIsMethodParameterFinal;
    property IsMethodParameterSynthetic[Index: Integer]: Boolean read GetIsMethodParameterSynthetic;
    property IsMethodParameterMandated[Index: Integer]: Boolean read GetIsMethodParameterMandated;
  end;

  // Field

  TField = class(TClassFileItem)
  private
    FFlags: Word;
    FNameIndex: Word;
    FDescriptorIndex: Word;
    FAttributes: TAttributeDynArray;
    function GetIsPublic: Boolean;
    function GetIsPrivate: Boolean;
    function GetIsProtected: Boolean;
    function GetIsEnum: Boolean;
    function GetIsFinal: Boolean;
    function GetIsStatic: Boolean;
    function GetIsSynthetic: Boolean;
    function GetIsTransient: Boolean;
    function GetIsVolatile: Boolean;
    function GetAttribute(Index: Integer): TAttribute;
    function GetAttributeCount: Integer;
    function GetDescriptor: TConstantUtf8;
    function GetName: TConstantUtf8;
    function GetIsConstant: Boolean;
    function GetIsIntegerConstant: Boolean;
    function GetIsLongConstant: Boolean;
    function GetIsFloatConstant: Boolean;
    function GetIsDoubleConstant: Boolean;
    function GetIsStringConstant: Boolean;
    function GetConstant: TConstantValueAttribute;
    function GetIntegerConstant: LongWord;
    function GetLongConstant: UInt64;
    function GetFloatConstant: Single;
    function GetDoubleConstant: Double;
    function GetStringConstant: TString;
  protected
    constructor Create(ClassFile: TClassFile; var Offset: Integer);
  public
    destructor Destroy; override;
    property Flags: Word read FFlags;
    property NameIndex: Word read FNameIndex;
    property Name: TConstantUtf8 read GetName;
    property DescriptorIndex: Word read FDescriptorIndex;
    property Descriptor: TConstantUtf8 read GetDescriptor;
    property AttributeCount: Integer read GetAttributeCount;
    property Attribute[Index: Integer]: TAttribute read GetAttribute;

    property IsPublic: Boolean read GetIsPublic;
    property IsPrivate: Boolean read GetIsPrivate;
    property IsProtected: Boolean read GetIsProtected;
    property IsStatic: Boolean read GetIsStatic;
    property IsFinal: Boolean read GetIsFinal;
    property IsVolatile: Boolean read GetIsVolatile;
    property IsTransient: Boolean read GetIsTransient;
    property IsSynthetic: Boolean read GetIsSynthetic;
    property IsEnum: Boolean read GetIsEnum;

    property IsConstant: Boolean read GetIsConstant;
    property IsIntegerConstant: Boolean read GetIsIntegerConstant;
    property IsLongConstant: Boolean read GetIsLongConstant;
    property IsFloatConstant: Boolean read GetIsFloatConstant;
    property IsDoubleConstant: Boolean read GetIsDoubleConstant;
    property IsStringConstant: Boolean read GetIsStringConstant;
    property Constant: TConstantValueAttribute read GetConstant;
    property IntegerContant: LongWord read GetIntegerConstant;
    property LongConstant: UInt64 read GetLongConstant;
    property FloatConstant: Single read GetFloatConstant;
    property DoubleConstant: Double read GetDoubleConstant;
    property StringConstant: TString read GetStringConstant;
  end;

  TFieldDynArray = array of TField;

  // Method

  TMethod = class(TClassFileItem)
  private
    FFlags: Word;
    FNameIndex: Word;
    FDescriptorIndex: Word;
    FAttributes: TAttributeDynArray;
    FCode: TCodeAttribute;
    FLocalVariables: TLocalVariableTableAttribute;
    FMethodParameters: TMethodParametersAttribute;
    FParameters: TIntegerDynArray;
    function GetIsAbstract: Boolean;
    function GetIsBridge: Boolean;
    function GetIsFinal: Boolean;
    function GetIsNative: Boolean;
    function GetIsPrivate: Boolean;
    function GetIsProtected: Boolean;
    function GetIsPublic: Boolean;
    function GetIsStatic: Boolean;
    function GetIsStrict: Boolean;
    function GetIsSynchronized: Boolean;
    function GetIsSynthetic: Boolean;
    function GetIsVarargs: Boolean;
    function GetDescriptor: TConstantUtf8;
    function GetName: TConstantUtf8;
    function GetAttribute(Index: Integer): TAttribute;
    function GetAttributeCount: Integer;
    function GetIsProcedure: Boolean;
    function GetParameter(Index: Integer): TSignature;
    function GetParameterCount: Integer;
    function GetResultType: TSignature;
    function GetParameterName(Index: Integer): TString;
    function GetParameterType(Index: Integer): TString;
    function GetCode: TCodeAttribute;
    function GetLocalVariables: TLocalVariableTableAttribute;
    function GetMethodParameters: TMethodParametersAttribute;
    function GetIsDeprecated: Boolean;
    function GetIsConstructor: Boolean;
    function GetParameters: TIntegerDynArray;
    procedure CheckParameterIndex(Index: Integer);
    property Parameters: TIntegerDynArray read GetParameters;
  protected
    constructor Create(ClassFile: TClassFile; var Offset: Integer);
  public
    destructor Destroy; override;
    property Flags: Word read FFlags;
    property NameIndex: Word read FNameIndex;
    property Name: TConstantUtf8 read GetName;
    property DescriptorIndex: Word read FDescriptorIndex;
    property Descriptor: TConstantUtf8 read GetDescriptor;
    property AttributeCount: Integer read GetAttributeCount;
    property Attribute[Index: Integer]: TAttribute read GetAttribute;
    property ParameterCount: Integer read GetParameterCount;
    property Parameter[Index: Integer]: TSignature read GetParameter;
    property ParameterName[Index: Integer]: TString read GetParameterName;
    property ParameterType[Index: Integer]: TString read GetParameterType;
    property ResultType: TSignature read GetResultType;
    property Code: TCodeAttribute read GetCode;
    property LocalVariables: TLocalVariableTableAttribute read GetLocalVariables;
    property MethodParameters: TMethodParametersAttribute read GetMethodParameters;

    property IsPublic: Boolean read GetIsPublic;
    property IsPrivate: Boolean read GetIsPrivate;
    property IsProtected: Boolean read GetIsProtected;
    property IsStatic: Boolean read GetIsStatic;
    property IsFinal: Boolean read GetIsFinal;
    property IsSynchronized: Boolean read GetIsSynchronized;
    property IsBridge: Boolean read GetIsBridge;
    property IsVarargs: Boolean read GetIsVarargs;
    property IsNative: Boolean read GetIsNative;
    property IsAbstract: Boolean read GetIsAbstract;
    property IsStrict: Boolean read GetIsStrict;
    property IsSynthetic: Boolean read GetIsSynthetic;

    property IsProcedure: Boolean read GetIsProcedure;
    property IsDeprecated: Boolean read GetIsDeprecated;
    property IsConstructor: Boolean read GetIsConstructor;
  end;

  TMethodDynArray = array of TMethod;

  // ClassFile

  TClassFile = class
  private
    FFileName: string;
    FData: TByteDynArray;
    FMagic: LongWord;
    FMinorVersion: Word;
    FMajorVersion: Word;
    FConstantPool: TConstantPoolItemDynArray;
    FFlags: Word;
    FThisClassIndex: Word;
    FSuperClassIndex: Word;
    FInterfaces: TWordDynArray;
    FFields: TFieldDynArray;
    FMethods: TMethodDynArray;
    FAttributes: TAttributeDynArray;
//    FIsInnerClass: Boolean;
    function GetByte(var Offset: Integer): Byte;
    function GetUInt16(var Offset: Integer): UInt16;
    function GetUInt32(var Offset: Integer): UInt32;
    function GetUInt64(var Offset: Integer): UInt64;
    procedure CreateConstantPool(var Offset: Integer);
    function GetIsPublic: Boolean;
    function GetIsFinal: Boolean;
    function GetSpecialSuper: Boolean;
    function GetIsInterface: Boolean;
    function GetIsAbstract: Boolean;
    function GetIsSynthetic: Boolean;
    function GetIsAnotation: Boolean;
    function GetIsEnum: Boolean;
    function GetField(Index: Integer): TField;
    function GetMethod(Index: Integer): TMethod;
    function GetAttribute(Index: Integer): TAttribute;
    function GetConstantPoolUtf8(Index: Integer): TConstantUtf8;
    function GetConstantPoolItem(Index: Integer): TConstantPoolItem;
    function GetConstantPoolClass(Index: Integer): TConstantClass;
    function GetConstantPoolNameAndType(Index: Integer): TConstantNameAndType;
    function GetAttributeCount: Integer;
    function GetFieldCount: Integer;
    function GetMethodCount: Integer;
    function GetInterface(Index: Integer): TConstantClass;
    function GetInterfaceCount: Integer;
    function GetSuperClass: TConstantClass;
    function GetThisClass: TConstantClass;
//    function GetIsInnerClass: Boolean;
  public
    constructor Create(const FileName: string; Data: TByteDynArray);
    destructor Destroy; override;
    property FileName: string read FFileName;
    property Magic: LongWord read FMagic;
    property MinorVersion: Word read FMinorVersion;
    property MajorVersion: Word read FMajorVersion;
    property ConstantPool: TConstantPoolItemDynArray read FConstantPool;
    property ConstantPoolItem[Index: Integer]: TConstantPoolItem read GetConstantPoolItem;
    property ConstantPoolUtf8[Index: Integer]: TConstantUtf8 read GetConstantPoolUtf8;
    property ConstantPoolClass[Index: Integer]: TConstantClass read GetConstantPoolClass;
    property ConstantPoolNameAndType[Index: Integer]: TConstantNameAndType read GetConstantPoolNameAndType;
    property Flags: Word read FFlags;
    property ThisClassIndex: Word read FThisClassIndex;
    property ThisClass: TConstantClass read GetThisClass;
    property SuperClassIndex: Word read FSuperClassIndex;
    property SuperClass: TConstantClass read GetSuperClass;
    property InterfaceCount: Integer read GetInterfaceCount;
    property Interface_[Index: Integer]: TConstantClass read GetInterface;
    property FieldCount: Integer read GetFieldCount;
    property Field[Index: Integer]: TField read GetField;
    property MethodCount: Integer read GetMethodCount;
    property Method[Index: Integer]: TMethod read GetMethod;
    property AttributeCount: Integer read GetAttributeCount;
    property Attribute[Index: Integer]: TAttribute read GetAttribute;

    property IsPublic: Boolean read GetIsPublic;
    property IsFinal: Boolean read GetIsFinal;
    property SpecialSuper: Boolean read GetSpecialSuper;
    property IsInterface: Boolean read GetIsInterface;
    property IsAbstract: Boolean read GetIsAbstract;
    property IsSynthetic: Boolean read GetIsSynthetic;
    property IsAnotation: Boolean read GetIsAnotation;
    property IsEnum: Boolean read GetIsEnum;

//    property IsInnerClass: Boolean read GetIsInnerClass;
  end;

implementation

uses SysUtils {$ifndef DXE3PLUS}, Math {$endif DXE3PLUS};

function SwapUInt16(Value: UInt16): UInt16;
begin
  Result := (Value shl 8) or (Value shr 8);
end;

function SwapUInt32(Value: UInt32): UInt32;
begin
  Result := (Value shl 24) or (Value shr 24) or ((Value and $FF00) shl 8) or ((Value and $FF0000) shr 8);
end;

function SwapUInt64(Value: UInt64): UInt64;
begin
  Result := (Value shl (7 * 8)) or ((Value and $FF00) shl (5 * 8)) or ((Value and $FF0000) shl (3 * 8)) or ((Value and $FF000000) shl 8) or
    ((Value shr (7 * 8)) and $FF) or ((Value shr 5 * 8) and $FF00) or ((Value shr (3 * 8)) and $FF0000) or ((Value shr 8) and $FF000000);
end;

procedure Check(Condition: Boolean; Message: string);
begin
  if not Condition then
    raise Exception.Create(Message);
end;

// TClassFileItem

constructor TClassFileItem.Create(ClassFile: TClassFile);
begin
  inherited Create;
  FClassFile := ClassFile;
end;

// TClassFile

constructor TClassFile.Create(const FileName: string; Data: TByteDynArray);
var InterfaceCount, FieldCount, MethodCount, AttributeCount, Offset, I: Integer;
// InnerClassesAttribute: TInnerClassesAttribute;
begin
  inherited Create;
  FFileName := FileName;
  FData := Data;
  Offset := 0;
  FMagic := GetUInt32(Offset);
  Check(FMagic = $CAFEBABE, FileName + ': Incorrect magic number');
  FMinorVersion := GetUInt16(Offset);
  FMajorVersion := GetUInt16(Offset);
  CreateConstantPool(Offset);
  FFlags := GetUInt16(Offset);
  FThisClassIndex := GetUInt16(Offset);
  FSuperClassIndex := GetUInt16(Offset);

  InterfaceCount := GetUInt16(Offset);
  SetLength(FInterfaces, InterfaceCount);
  for I := 0 to InterfaceCount - 1 do
    FInterfaces[I] := GetUInt16(Offset);

  FieldCount := GetUInt16(Offset);
  SetLength(FFields, FieldCount);
  for I := 0 to FieldCount - 1 do
    FFields[I] := TField.Create(Self, Offset);

  MethodCount := GetUInt16(Offset);
  SetLength(FMethods, MethodCount);
  for I := 0 to MethodCount - 1 do
    FMethods[I] := TMethod.Create(Self, Offset);

  AttributeCount := GetUInt16(Offset);
  SetLength(FAttributes, AttributeCount);
  for I := 0 to AttributeCount - 1 do
  begin
    FAttributes[I] := TAttribute.CreateAttribute(Self, Offset);
{
    if FAttributes[I] is TInnerClassesAttribute then
    begin
      InnerClassesAttribute := TInnerClassesAttribute(FAttributes[I]);
      for J := InnerClassesAttribute.InnerClassCount - 1 downto 0 do
        if InnerClassesAttribute.InnerClass[J].InnerClassIndex = FThisClassIndex then
        begin
          FIsInnerClass := True;
          Break;
        end;
    end;
}
  end;
end;

destructor TClassFile.Destroy;
var
  I: Integer;
begin
  for I := 0 to Length(FConstantPool) - 1 do
    FreeAndNil(FConstantPool[I]);
  FConstantPool := nil;

  FInterfaces := nil;

  for I := 0 to Length(FFields) - 1 do
    FreeAndNil(FFields[I]);
  FFields := nil;

  for I := 0 to Length(FMethods) - 1 do
    FreeAndNil(FMethods[I]);
  FMethods := nil;

  for I := 0 to Length(FAttributes) - 1 do
    FreeAndNil(FAttributes[I]);
  FAttributes := nil;

  inherited;
end;

function TClassFile.GetByte(var Offset: Integer): Byte;
begin
  Check((Offset >= 0) and (Offset + SizeOf(Byte) <= Length(FData)), FileName + ': Error reading byte');
  Result := FData[Offset];
  Inc(Offset, SizeOf(Byte));
end;

function TClassFile.GetUInt16(var Offset: Integer): UInt16;
begin
  Check((Offset >= 0) and (Offset + SizeOf(Word) <= Length(FData)), FileName + ': Error reading word');
  Result := SwapUInt16(PWord(@FData[Offset])^);
  Inc(Offset, SizeOf(Word));
end;

function TClassFile.GetUInt32(var Offset: Integer): UInt32;
begin
  Check((Offset >= 0) and (Offset + SizeOf(LongWord) <= Length(FData)), FileName + ': Error reading long word');
  Result := SwapUInt32(PLongWord(@FData[Offset])^);
  Inc(Offset, SizeOf(LongWord));
end;

function TClassFile.GetUInt64(var Offset: Integer): UInt64;
begin
  Check((Offset >= 0) and (Offset + SizeOf(UInt64) <= Length(FData)), FileName + ': Error reading quad word');
  Result := SwapUInt64(PUInt64(@FData[Offset])^);
  Inc(Offset, SizeOf(UInt64));
end;

procedure TClassFile.CreateConstantPool(var Offset: Integer);
var
  ConstantPoolCount, I: Integer;
  Tag: Byte;
begin
  ConstantPoolCount := GetUInt16(Offset) - 1;
  SetLength(FConstantPool, ConstantPoolCount);
  I := 0;
  while I < ConstantPoolCount do
  begin
    Tag := GetByte(Offset);
    case Tag of
      JVM_CONSTANT_Utf8:
        FConstantPool[I] := TConstantUtf8.Create(Self, Offset);
      JVM_CONSTANT_Integer:
        FConstantPool[I] := TConstantInteger.Create(Self, Offset);
      JVM_CONSTANT_Float:
        FConstantPool[I] := TConstantFloat.Create(Self, Offset);
      JVM_CONSTANT_Long:
      begin
        FConstantPool[I] := TConstantLong.Create(Self, Offset);
        Inc(I);
      end;
      JVM_CONSTANT_Double:
      begin
        FConstantPool[I] := TConstantDouble.Create(Self, Offset);
        Inc(I);
      end;
      JVM_CONSTANT_Class:
        FConstantPool[I] := TConstantClass.Create(Self, Offset);
      JVM_CONSTANT_String:
        FConstantPool[I] := TConstantString.Create(Self, Offset);
      JVM_CONSTANT_Fieldref:
        FConstantPool[I] := TConstantFieldRef.Create(Self, Offset);
      JVM_CONSTANT_Methodref:
        FConstantPool[I] := TConstantMethodRef.Create(Self, Offset);
      JVM_CONSTANT_InterfaceMethodref:
        FConstantPool[I] := TConstantInterfaceMethodRef.Create(Self, Offset);
      JVM_CONSTANT_NameAndType:
        FConstantPool[I] := TConstantNameAndType.Create(Self, Offset);
      JVM_CONSTANT_MethodHandle:
        FConstantPool[I] := TConstantMethodHandle.Create(Self, Offset);
      JVM_CONSTANT_MethodType:
        FConstantPool[I] := TConstantMethodType.Create(Self, Offset);
      JVM_CONSTANT_InvokeDynamic:
        FConstantPool[I] := TConstantInvokeDynamic.Create(Self, Offset);
      else
        Check(False, FileName + ': Incorrect constant pool item');
    end;
    Inc(I);
  end;
end;

function TClassFile.GetConstantPoolItem(Index: Integer): TConstantPoolItem;
begin
  Check((Index >= 0) and (Index < Length(ConstantPool)), FileName + ': Incorrect constant pool index');
  Result := ConstantPool[Index];
end;

function TClassFile.GetConstantPoolUtf8(Index: Integer): TConstantUtf8;
begin
  Check(ConstantPoolItem[Index] is TConstantUtf8, FileName + ': UTF-8 constant pool item expected');
  Result := ConstantPoolItem[Index] as TConstantUtf8;
end;

function TClassFile.GetConstantPoolClass(Index: Integer): TConstantClass;
begin
  Check(ConstantPoolItem[Index] is TConstantClass, FileName + ': Class constant pool item expected');
  Result := ConstantPoolItem[Index] as TConstantClass;
end;

function TClassFile.GetConstantPoolNameAndType(Index: Integer): TConstantNameAndType;
begin
  Check(ConstantPoolItem[Index] is TConstantNameAndType, FileName + ': Name and type constant pool item expected');
  Result := ConstantPoolItem[Index] as TConstantNameAndType;
end;

function TClassFile.GetInterfaceCount: Integer;
begin
  Result := Length(FInterfaces);
end;

function TClassFile.GetInterface(Index: Integer): TConstantClass;
begin
  Check((Index >= 0) and (Index < InterfaceCount), FileName + ': Incorrect interface index');
  Result := ConstantPoolClass[FInterfaces[Index] - 1];
end;

function TClassFile.GetThisClass: TConstantClass;
begin
  Result := ConstantPoolClass[FThisClassIndex - 1];
end;

function TClassFile.GetSuperClass: TConstantClass;
begin
  if FSuperClassIndex = 0 then
    Result := nil
  else
    Result := ConstantPoolClass[FSuperClassIndex - 1];
end;

function TClassFile.GetFieldCount: Integer;
begin
  Result := Length(FFields);
end;

function TClassFile.GetField(Index: Integer): TField;
begin
  Check((Index >= 0) and (Index < FieldCount), FileName + ': Incorrect field index');
  Result := FFields[Index];
end;

function TClassFile.GetMethodCount: Integer;
begin
  Result := Length(FMethods);
end;

function TClassFile.GetMethod(Index: Integer): TMethod;
begin
  Check((Index >= 0) and (Index < MethodCount), FileName + ': Incorrect method index');
  Result := FMethods[Index];
end;

function TClassFile.GetAttributeCount: Integer;
begin
  Result := Length(FAttributes);
end;

function TClassFile.GetAttribute(Index: Integer): TAttribute;
begin
  Check((Index >= 0) and (Index < AttributeCount), FileName + ': Incorrect attribute index');
  Result := FAttributes[Index];
end;

function TClassFile.GetIsPublic: Boolean;
begin
  Result := (Flags and JVM_ACC_PUBLIC) <> 0;
end;

function TClassFile.GetIsFinal: Boolean;
begin
  Result := (Flags and JVM_ACC_FINAL) <> 0;
end;

function TClassFile.GetSpecialSuper: Boolean;
begin
  Result := (Flags and JVM_ACC_SUPER) <> 0;
end;

function TClassFile.GetIsInterface: Boolean;
begin
  Result := (Flags and JVM_ACC_INTERFACE) <> 0;
end;

function TClassFile.GetIsAbstract: Boolean;
begin
  Result := (Flags and JVM_ACC_ABSTRACT) <> 0;
end;

function TClassFile.GetIsSynthetic: Boolean;
begin
  Result := (Flags and JVM_ACC_SYNTHETIC) <> 0;
end;

function TClassFile.GetIsAnotation: Boolean;
begin
  Result := (Flags and JVM_ACC_ANNOTATION) <> 0;
end;

function TClassFile.GetIsEnum: Boolean;
begin
  Result := (Flags and JVM_ACC_ENUM) <> 0;
end;

{
function TClassFile.GetIsInnerClass: Boolean;
begin
  Result := FIsInnerClass;
end;
}

// TField

constructor TField.Create(ClassFile: TClassFile; var Offset: Integer);
var AttributeCount, I: Integer;
begin
  inherited Create(ClassFile);
  FFlags := ClassFile.GetUInt16(Offset);
  FNameIndex := ClassFile.GetUInt16(Offset);
  FDescriptorIndex := ClassFile.GetUInt16(Offset);
  AttributeCount := ClassFile.GetUInt16(Offset);
  SetLength(FAttributes, AttributeCount);
  for I := 0 to Length(FAttributes) - 1 do
    FAttributes[I] := TAttribute.CreateAttribute(ClassFile, Offset);
end;

destructor TField.Destroy;
var
  I: Integer;
begin
  for I := 0 to Length(FAttributes) - 1 do
    FreeAndNil(FAttributes[I]);
  FAttributes := nil;

  inherited;
end;

function TField.GetName: TConstantUtf8;
begin
  Result := ClassFile.ConstantPoolUtf8[NameIndex - 1];
end;

function TField.GetDescriptor: TConstantUtf8;
begin
  Result := ClassFile.ConstantPoolUtf8[DescriptorIndex - 1];
end;

function TField.GetAttributeCount: Integer;
begin
  Result := Length(FAttributes);
end;

function TField.GetAttribute(Index: Integer): TAttribute;
begin
  Check((Index >= 0) and (Index < Length(FAttributes)), ClassFile.FileName + ': Incorrect attribute index');
  Result := FAttributes[Index];
end;

function TField.GetIsPublic: Boolean;
begin
  Result := (Flags and JVM_ACC_PUBLIC) <> 0;
end;

function TField.GetIsPrivate: Boolean;
begin
  Result := (Flags and JVM_ACC_PRIVATE) <> 0;
end;

function TField.GetIsProtected: Boolean;
begin
  Result := (Flags and JVM_ACC_PROTECTED) <> 0;
end;

function TField.GetIsStatic: Boolean;
begin
  Result := (Flags and JVM_ACC_STATIC) <> 0;
end;

function TField.GetIsFinal: Boolean;
begin
  Result := (Flags and JVM_ACC_FINAL) <> 0;
end;

function TField.GetIsVolatile: Boolean;
begin
  Result := (Flags and JVM_ACC_VOLATILE) <> 0;
end;

function TField.GetIsTransient: Boolean;
begin
  Result := (Flags and JVM_ACC_TRANSIENT) <> 0;
end;

function TField.GetIsSynthetic: Boolean;
begin
  Result := (Flags and JVM_ACC_SYNTHETIC) <> 0;
end;

function TField.GetIsEnum: Boolean;
begin
  Result := (Flags and JVM_ACC_ENUM) <> 0;
end;

function TField.GetIsConstant: Boolean;
begin
  Result := Constant <> nil;
end;

function TField.GetIsIntegerConstant: Boolean;
begin
  Result := IsConstant and (Constant.Value is TConstantInteger);
end;

function TField.GetIsFloatConstant: Boolean;
begin
  Result := IsConstant and (Constant.Value is TConstantFloat);
end;

function TField.GetIsLongConstant: Boolean;
begin
  Result := IsConstant and (Constant.Value is TConstantLong);
end;

function TField.GetIsDoubleConstant: Boolean;
begin
  Result := IsConstant and (Constant.Value is TConstantDouble);
end;

function TField.GetIsStringConstant: Boolean;
begin
  Result := IsConstant and (Constant.Value is TConstantString);
end;

function TField.GetConstant: TConstantValueAttribute;
var I: Integer;
begin
  Result := nil;

  for I := 0 to AttributeCount - 1 do
    if Attribute[I] is TConstantValueAttribute then
    begin
      Result := TConstantValueAttribute(Attribute[I]);
      Break;
    end;
end;

function TField.GetIntegerConstant: LongWord;
begin
  Check(IsIntegerConstant, ClassFile.FileName + ': Not an integer constant');

  Result := TConstantInteger(Constant.Value).Value;
end;

function TField.GetLongConstant: UInt64;
begin
  Check(IsLongConstant, ClassFile.FileName + ': Not a long constant');

  Result := TConstantLong(Constant.Value).Value;
end;

function TField.GetFloatConstant: Single;
begin
  Check(IsFloatConstant, ClassFile.FileName + ': Not a float constant');

  Result := TConstantFloat(Constant.Value).Value;
end;

function TField.GetDoubleConstant: Double;
begin
  Check(IsDoubleConstant, ClassFile.FileName + ': Not a double constant');

  Result := TConstantDouble(Constant.Value).Value;
end;

function TField.GetStringConstant: TString;
begin
  Check(IsStringConstant, ClassFile.FileName + ': Not a string constant');

  Result := TConstantString(Constant.Value).StringValue;
end;

// TMethod

constructor TMethod.Create(ClassFile: TClassFile; var Offset: Integer);
var AttributeCount, I: Integer;
begin
  inherited Create(ClassFile);
  FFlags := ClassFile.GetUInt16(Offset);
  FNameIndex := ClassFile.GetUInt16(Offset);
  FDescriptorIndex := ClassFile.GetUInt16(Offset);
  AttributeCount := ClassFile.GetUInt16(Offset);
  SetLength(FAttributes, AttributeCount);
  for I := 0 to Length(FAttributes) - 1 do
    FAttributes[I] := TAttribute.CreateAttribute(ClassFile, Offset);
end;

destructor TMethod.Destroy;
var
  I: Integer;
begin
  for I := 0 to Length(FAttributes) - 1 do
    FreeAndNil(FAttributes[I]);
  FAttributes := nil;

  inherited;
end;

function TMethod.GetName: TConstantUtf8;
begin
  Result := ClassFile.ConstantPoolUtf8[FNameIndex - 1];
end;

function TMethod.GetDescriptor: TConstantUtf8;
begin
  Result := ClassFile.ConstantPoolUtf8[FDescriptorIndex - 1];
end;

function TMethod.GetAttributeCount: Integer;
begin
  Result := Length(FAttributes);
end;

function TMethod.GetAttribute(Index: Integer): TAttribute;
begin
  Check((Index >= 0) and (Index < AttributeCount), ClassFile.FileName + ': Incorrect attribute index');
  Result := FAttributes[Index];
end;

function TMethod.GetIsPublic: Boolean;
begin
  Result := (Flags and JVM_ACC_PUBLIC) <> 0;
end;

function TMethod.GetIsPrivate: Boolean;
begin
  Result := (Flags and JVM_ACC_PRIVATE) <> 0;
end;

function TMethod.GetIsProtected: Boolean;
begin
  Result := (Flags and JVM_ACC_PROTECTED) <> 0;
end;

function TMethod.GetIsStatic: Boolean;
begin
  Result := (Flags and JVM_ACC_STATIC) <> 0;
end;

function TMethod.GetIsFinal: Boolean;
begin
  Result := (Flags and JVM_ACC_FINAL) <> 0;
end;

function TMethod.GetIsSynchronized: Boolean;
begin
  Result := (Flags and JVM_ACC_SYNCHRONIZED) <> 0;
end;

function TMethod.GetIsBridge: Boolean;
begin
  Result := (Flags and JVM_ACC_BRIDGE) <> 0;
end;

function TMethod.GetIsVarargs: Boolean;
begin
  Result := (Flags and JVM_ACC_VARARGS) <> 0;
end;

function TMethod.GetIsNative: Boolean;
begin
  Result := (Flags and JVM_ACC_NATIVE) <> 0;
end;

function TMethod.GetIsAbstract: Boolean;
begin
  Result := (Flags and JVM_ACC_ABSTRACT) <> 0;
end;

function TMethod.GetIsStrict: Boolean;
begin
  Result := (Flags and JVM_ACC_STRICT) <> 0;
end;

function TMethod.GetIsSynthetic: Boolean;
begin
  Result := (Flags and JVM_ACC_SYNTHETIC) <> 0;
end;

function TMethod.GetIsProcedure: Boolean;
begin
  Result := Descriptor.MethodSignature.ResultType = nil;
end;

function TMethod.GetParameterCount: Integer;
begin
  Result := Descriptor.MethodSignature.ParameterCount;
end;

function TMethod.GetParameter(Index: Integer): TSignature;
begin
  Result := Descriptor.MethodSignature.Parameter[Index];
end;

function TMethod.GetResultType: TSignature;
begin
  Result := Descriptor.MethodSignature.ResultType;
end;

function TMethod.GetCode: TCodeAttribute;
var I: Integer;
begin
  if FCode = nil then
  begin
    for I := 0 to AttributeCount - 1 do
      if Attribute[I] is TCodeAttribute then
      begin
        FCode := TCodeAttribute(Attribute[I]);
        Break;
      end;
  end;
  Result := FCode;
end;

function TMethod.GetLocalVariables: TLocalVariableTableAttribute;
var
  CodeAttribute: TCodeAttribute;
  I: Integer;
begin
  if FLocalVariables = nil then
  begin
    CodeAttribute := Code;
    if CodeAttribute <> nil then
      for I := 0 to CodeAttribute.AttributeCount - 1 do
        if CodeAttribute.Attribute[I] is TLocalVariableTableAttribute then
        begin
          FLocalVariables := TLocalVariableTableAttribute(CodeAttribute.Attribute[I]);
          Break;
        end;
  end;
  Result := FLocalVariables;
end;

function TMethod.GetMethodParameters: TMethodParametersAttribute;
var I: Integer;
begin
  if FMethodParameters = nil then
    for I := 0 to AttributeCount - 1 do
      if Attribute[I] is TMethodParametersAttribute then
      begin
        FMethodParameters := TMethodParametersAttribute(Attribute[I]);
        Break;
      end;
  Result := FMethodParameters;
end;

procedure TMethod.CheckParameterIndex(Index: Integer);
begin
  Check((Index >= 0) and (Index < ParameterCount), ClassFile.FileName + ': Incorrect parameter index');
end;

function TMethod.GetParameters: TIntegerDynArray;
var
  Variables: TLocalVariableTableAttribute;
  Locals: TLocalVariableDynArray;
  I, J, Frame: Integer;
begin
  Locals := nil; // to avoid warning
  if FParameters = nil then
  begin
    Variables := LocalVariables;
    if Variables <> nil then
    begin
      Locals := Variables.FLocalVariables;
      SetLength(FParameters, ParameterCount);

      Frame := 0;
      if not IsStatic then
        Inc(Frame); // skip 'this'

      for I := 0 to Length(FParameters) - 1 do
      begin
        FParameters[I] := -1;
        for J := Variables.LocalVariableCount - 1 downto 0 do
          if Locals[J].Index = Frame then
          begin
            FParameters[I] := J;
            Break;
          end;

        Inc(Frame);
        if (Parameter[I] is TBaseSignature) and (TBaseSignature(Parameter[I]).BaseType in [btLong, btDouble]) then
          Inc(Frame);
      end;
    end;
  end;
  Result := FParameters;
end;

function TMethod.GetParameterName(Index: Integer): TString;
begin
  CheckParameterIndex(Index);

  Result := '';
  if MethodParameters <> nil then
    Result := MethodParameters.MethodParameterName[Index];
  if Result = '' then
    if Parameters <> nil then
      if Parameters[Index] <> -1 then
        Result := LocalVariables.LocalVariableName[Parameters[Index]];
end;

function TMethod.GetParameterType(Index: Integer): TString;
begin
  CheckParameterIndex(Index);

  Result := '';
  if Parameters <> nil then
    if Parameters[Index] <> -1 then
      Result := LocalVariables.LocalVariableType[Parameters[Index]];
end;

function TMethod.GetIsDeprecated: Boolean;
var I: Integer;
begin
  Result := False;

  for I := 0 to AttributeCount - 1 do
    if Attribute[I] is TDeprecatedAttribute then
    begin
      Result := True;
      Break;
    end;
end;

function TMethod.GetIsConstructor: Boolean;
begin
  Result := Name.Value = '<init>';
end;

// TConstantPoolItem

constructor TConstantPoolItem.Create(ClassFile: TClassFile);
begin
  inherited Create(ClassFile);
end;

// TConstantClass

constructor TConstantClass.Create(ClassFile: TClassFile; var Offset: Integer);
begin
  inherited Create(ClassFile);
  FNameIndex := ClassFile.GetUInt16(Offset);
end;

function TConstantClass.GetName: TString;
begin
  Result := ClassFile.ConstantPoolUtf8[NameIndex - 1].Value;
end;

// TConstantInteger

constructor TConstantInteger.Create(ClassFile: TClassFile; var Offset: Integer);
begin
  inherited Create(ClassFile);
  FValue := ClassFile.GetUInt32(Offset);
end;

// TConstantFloat

constructor TConstantFloat.Create(ClassFile: TClassFile; var Offset: Integer);
var
  Bytes: LongWord;
begin
  inherited Create(ClassFile);
  Bytes := ClassFile.GetUInt32(Offset);
  FValue := PSingle(@Bytes)^;
end;

// TConstantFieldRef

constructor TConstantFieldRef.Create(ClassFile: TClassFile; var Offset: Integer);
begin
  inherited Create(ClassFile);
  FClassIndex := ClassFile.GetUInt16(Offset);
  FNameAndTypeIndex := ClassFile.GetUInt16(Offset);
end;

function TConstantFieldRef.GetClassValue: TString;
begin
  Result := ClassFile.ConstantPoolClass[ClassIndex - 1].Name;
end;

function TConstantFieldRef.GetName: TString;
begin
  Result := ClassFile.ConstantPoolNameAndType[NameAndTypeIndex - 1].Name;
end;

function TConstantFieldRef.GetTypeName: TString;
begin
  Result := ClassFile.ConstantPoolNameAndType[NameAndTypeIndex - 1].Descriptor;
end;

function TConstantFieldRef.GetTypeSignature: TSignature;
begin
  Result := ClassFile.ConstantPoolNameAndType[NameAndTypeIndex - 1].Signature;
  Check(not(Result is TMethodSignature), ClassFile.FileName + ': Unexpected method signature');
end;

// TConstantMethodRef

constructor TConstantMethodRef.Create(ClassFile: TClassFile; var Offset: Integer);
begin
  inherited Create(ClassFile);
  FClassIndex := ClassFile.GetUInt16(Offset);
  FNameAndTypeIndex := ClassFile.GetUInt16(Offset);
end;

function TConstantMethodRef.GetClassValue: TString;
begin
  Result := ClassFile.ConstantPoolClass[ClassIndex - 1].Name;
end;

function TConstantMethodRef.GetName: TString;
begin
  Result := ClassFile.ConstantPoolNameAndType[NameAndTypeIndex - 1].Name;
end;

function TConstantMethodRef.GetTypeName: TString;
begin
  Result := ClassFile.ConstantPoolNameAndType[NameAndTypeIndex - 1].Descriptor;
end;

function TConstantMethodRef.GetTypeSignature: TSignature;
begin
  Result := ClassFile.ConstantPoolNameAndType[NameAndTypeIndex - 1].Signature;
  Check(Result is TMethodSignature, ClassFile.FileName + ': Method signature expected');
end;

// TConstantInterfaceMethodRef

constructor TConstantInterfaceMethodRef.Create(ClassFile: TClassFile; var Offset: Integer);
begin
  inherited Create(ClassFile);
  FClassIndex := ClassFile.GetUInt16(Offset);
  FNameAndTypeIndex := ClassFile.GetUInt16(Offset);
end;

// TConstantString

constructor TConstantString.Create(ClassFile: TClassFile; var Offset: Integer);
begin
  inherited Create(ClassFile);
  FStringIndex := ClassFile.GetUInt16(Offset);
end;

function TConstantString.GetStringValue: TString;
begin
  Result := ClassFile.ConstantPoolUtf8[StringIndex - 1].Value;
end;

// TConstantLong

constructor TConstantLong.Create(ClassFile: TClassFile; var Offset: Integer);
begin
  inherited Create(ClassFile);
  FValue := ClassFile.GetUInt64(Offset);
end;

// TConstantDouble

constructor TConstantDouble.Create(ClassFile: TClassFile; var Offset: Integer);
{$ifndef D6PLUS}
const
  Infinity =  1.0 / 0.0;
  NegInfinity = -1.0 / 0.0;
  NaN =  0.0 / 0.0;
{$endif D6PLUS}
var
  Bytes: UInt64;
begin
  inherited Create(ClassFile);
  Bytes := ClassFile.GetUInt64(Offset);
  if Bytes = { $7ff0000000000000} $7FF0000000003C00 then
{$ifdef DXE3PLUS}
    FValue := Double.PositiveInfinity
{$else}
    FValue := Infinity
{$endif DXE3PLUS}
  else if Bytes = { $fff0000000000000} $FFF0000000003C00 then
{$ifdef DXE3PLUS}
    FValue := Double.NegativeInfinity
{$else}
    FValue := NegInfinity
{$endif DXE3PLUS}
  else if (Bytes >= $7ff0000000000001) and (Bytes <= $7fffffffffffffff) then
{$ifdef DXE3PLUS}
    FValue := Double.NaN
{$else}
    FValue := NaN
{$endif DXE3PLUS}
  else if (Bytes >= $fff0000000000001) and (Bytes <= $ffffffffffffffff) then
{$ifdef DXE3PLUS}
    FValue := Double.NaN
{$else}
    FValue := NaN
{$endif DXE3PLUS}
  else
    FValue := PDouble(@Bytes)^;
end;

// TConstantUtf8

constructor TConstantUtf8.Create(ClassFile: TClassFile; var Offset: Integer);
var
  Utf8Length: Word;
  Byte1, Byte2, Byte3: Byte;
  EndOffset, Index: Integer;
begin
  inherited Create(ClassFile);
  Utf8Length := ClassFile.GetUInt16(Offset);
  EndOffset := Offset + Utf8Length;

  SetLength(FValue, Utf8Length);
  Index := 0;
  while Offset < EndOffset do
  begin
    Byte1 := ClassFile.GetByte(Offset);
    CheckByte(Byte1);
    if (Byte1 and $80) = 0 then
    begin
      Inc(Index);
      FValue[Index] := WideChar(Byte1);
    end
    else
    begin
      Byte2 := ClassFile.GetByte(Offset);
      CheckByte(Byte2);
      if ((Byte1 and $E0) = $C0) and ((Byte2 and $C0) = $80) then
      begin
        Inc(Index);
        FValue[Index] := WideChar(((Byte1 and $1F) shl 6) + (Byte2 and $3F))
      end
      else
      begin
        Byte3 := ClassFile.GetByte(Offset);
        CheckByte(Byte3);
        if ((Byte1 and $F0) = $E0) and ((Byte2 and $C0) = $80) and ((Byte3 and $C0) = $80) then
        begin
          Inc(Index);
          FValue[Index] := WideChar(((Byte1 and $0F) shl 12) + ((Byte2 and $3F) shl 6) + (Byte3 and $3F))
        end
        else
          Check(False, ClassFile.FileName + ': Unsupported or incorrect character in UTF-8 constant pool');
      end;
    end;
  end;
  SetLength(FValue, Index);
end;

procedure TConstantUtf8.CheckByte(Value: Byte);
begin
  Check((Value > 0) and (Value < $F0), ClassFile.FileName + ': Incorrect value in UTF-8 constant pool');
end;

function TConstantUtf8.GetSignature: TSignature;
begin
  if FSignature = nil then
    FSignature := TSignature.CreateSignature(ClassFile, Value);
  Result := FSignature;
end;

function TConstantUtf8.GetMethodSignature: TMethodSignature;
begin
  Check(Signature is TMethodSignature, ClassFile.FileName + ': Method signature expected');
  Result := Signature as TMethodSignature;
end;

// TConstantNameAndType

constructor TConstantNameAndType.Create(ClassFile: TClassFile; var Offset: Integer);
begin
  inherited Create(ClassFile);
  FNameIndex := ClassFile.GetUInt16(Offset);
  FDescriptorIndex := ClassFile.GetUInt16(Offset);
end;

function TConstantNameAndType.GetName: TString;
begin
  Result := ClassFile.ConstantPoolUtf8[NameIndex - 1].Value;
end;

function TConstantNameAndType.GetDescriptor: TString;
begin
  Result := ClassFile.ConstantPoolUtf8[DescriptorIndex - 1].Value;
end;

function TConstantNameAndType.GetSignature: TSignature;
begin
  Result := ClassFile.ConstantPoolUtf8[DescriptorIndex - 1].Signature;
end;

// TConstantMethodHandle

constructor TConstantMethodHandle.Create(ClassFile: TClassFile; var Offset: Integer);
begin
  inherited Create(ClassFile);
  FReferenceKind := ClassFile.GetByte(Offset);
  FReferenceIndex := ClassFile.GetUInt16(Offset);
end;

// TConstantMethodType

constructor TConstantMethodType.Create(ClassFile: TClassFile; var Offset: Integer);
begin
  inherited Create(ClassFile);
  FDescriptorIndex := ClassFile.GetUInt16(Offset);
end;

function TConstantMethodType.GetDescriptor: TString;
begin
  Result := ClassFile.ConstantPoolUtf8[DescriptorIndex - 1].Value;
end;

// TConstantInvokeDynamic

constructor TConstantInvokeDynamic.Create(ClassFile: TClassFile; var Offset: Integer);
begin
  inherited Create(ClassFile);
  FBootstrapMethodAttrIndex := ClassFile.GetUInt16(Offset);
  FNameAndTypeIndex := ClassFile.GetUInt16(Offset);
end;

// TBaseSignature

constructor TBaseSignature.Create(ClassFile: TClassFile; BaseType: TBaseType);
begin
  inherited Create(ClassFile);
  FBaseType := BaseType;
end;

// TObjectSignature

constructor TObjectSignature.Create(ClassFile: TClassFile; const Name: TString);
begin
  inherited Create(ClassFile);
  FName := Name;
end;

// TArraySignature

constructor TArraySignature.Create(ClassFile: TClassFile; Dimension: Integer; ItemSignature: TSignature);
begin
  inherited Create(ClassFile);
  FDimension := Dimension;
  FItemSignature := ItemSignature;
end;

destructor TArraySignature.Destroy;
begin
  FreeAndNil(FItemSignature);
  inherited;
end;

// TMethodSignature

constructor TMethodSignature.Create(ClassFile: TClassFile; Parameters: TSignatureDynArray; ResultType: TSignature);
begin
  inherited Create(ClassFile);
  FParameters := Parameters;
  FResultType := ResultType;
end;

destructor TMethodSignature.Destroy;
var
  I: Integer;
begin
  for I := 0 to Length(FParameters) - 1 do
    FreeAndNil(FParameters[I]);
  FreeAndNil(FResultType);
  inherited;
end;

function TMethodSignature.GetParameterCount: Integer;
begin
  Result := Length(FParameters);
end;

function TMethodSignature.GetParameter(Index: Integer): TSignature;
begin
  Check((Index >= 0) and (Index < ParameterCount), ClassFile.FileName + ': Incorrect parameter index in method');
  Result := FParameters[Index];
end;

// TSignature

function GetBaseType(Value: WideChar): TBaseType;
begin
  case Value of
    'B':
      Result := btByte;
    'C':
      Result := btChar;
    'D':
      Result := btDouble;
    'F':
      Result := btFloat;
    'I':
      Result := btInt;
    'J':
      Result := btLong;
    'S':
      Result := btShort;
    'Z':
      Result := btBoolean;
  else
    Result := btUnknown;
  end;
end;

class function TSignature.CreateSignature(ClassFile: TClassFile; const Signature: TString; var Index: Integer): TSignature;
var
  BaseType: TBaseType;
  StartIndex, Dimension: Integer;
  Parameters: TSignatureDynArray;
begin
  Result := nil;
  Check((Index >= 1) and (Index <= Length(Signature)), ClassFile.FileName + ': Incorrect signature index');

  case Signature[Index] of
    '(':
      begin
        Check(Index = 1, ClassFile.FileName + ': Unexpected method signature');

        Inc(Index);
        while (Index <= Length(Signature)) and (Signature[Index] <> ')') do
        begin
          SetLength(Parameters, Length(Parameters) + 1);
          Parameters[Length(Parameters) - 1] := CreateSignature(ClassFile, Signature, Index);
        end;

        Check((Index <= Length(Signature)) and (Signature[Index] = ')'), ClassFile.FileName + ': Incorrect method signature');
        Inc(Index);
        if (Index <= Length(Signature)) and (Signature[Index] = 'V') then
        begin
          Inc(Index);
          Result := TMethodSignature.Create(ClassFile, Parameters, nil)
        end
        else
          Result := TMethodSignature.Create(ClassFile, Parameters, CreateSignature(ClassFile, Signature, Index));
      end;

    'L':
      begin
        StartIndex := Index;
        repeat
          Inc(Index);
        until (Index > Length(Signature)) or (Signature[Index] = ';');
        if (Index > Length(Signature)) then
          Check(False, ClassFile.FileName + ': Incorrect object type signature')
        else
        begin
          Inc(Index);
          Result := TObjectSignature.Create(ClassFile, Copy(Signature, StartIndex + 1, Index - StartIndex - 2));
        end;
      end;

    '[':
      begin
        Dimension := 0;
        repeat
          Inc(Index);
          Inc(Dimension);
        until (Index > Length(Signature)) or (Signature[Index] <> '[');
        Result := TArraySignature.Create(ClassFile, Dimension, CreateSignature(ClassFile, Signature, Index));
      end;

  else
    begin
      BaseType := GetBaseType(Signature[Index]);
      if BaseType = btUnknown then
        Check(False, ClassFile.FileName + ': Unknown signature ' + Signature[Index])
      else
      begin
        Inc(Index);
        Result := TBaseSignature.Create(ClassFile, BaseType);
      end
    end
  end;
end;

class function TSignature.CreateSignature(ClassFile: TClassFile; const Signature: TString): TSignature;
var
  Index: Integer;
begin
  Index := 1;
  Result := CreateSignature(ClassFile, Signature, Index);
  Check(Index = Length(Signature) + 1, ClassFile.FileName + ': Incorrect signature');
end;

constructor TSignature.Create(ClassFile: TClassFile);
begin
  inherited Create(ClassFile);
end;

// TAttribute

class function TAttribute.CreateAttribute(ClassFile: TClassFile; var Offset: Integer): TAttribute;
var
  Start: Integer;
  NameIndex: Word;
  Name: TString;
  Length: Integer;
begin
  NameIndex := ClassFile.GetUInt16(Offset);
  Name := ClassFile.ConstantPoolUtf8[NameIndex - 1].Value;
  Length := ClassFile.GetUInt32(Offset);
  Start := Offset;

  if Name = 'ConstantValue' then
    Result := TConstantValueAttribute.Create(ClassFile, Offset)
  else if Name = 'SourceFile' then
    Result := TSourceFileAttribute.Create(ClassFile, Offset)
  else if Name = 'Deprecated' then
    Result := TDeprecatedAttribute.Create(ClassFile, Offset)
  else if Name = 'Signature' then
    Result := TSignatureAttribute.Create(ClassFile, Offset)
  else if Name = 'Code' then
    Result := TCodeAttribute.Create(ClassFile, Offset)
  else if Name = 'InnerClasses' then
    Result := TInnerClassesAttribute.Create(ClassFile, Offset)
  else if Name = 'LineNumberTable' then
    Result := TLineNumberTableAttribute.Create(ClassFile, Offset)
  else if Name = 'LocalVariableTable' then
    Result := TLocalVariableTableAttribute.Create(ClassFile, Offset)
  else if Name = 'MethodParameters' then
    Result := TMethodParametersAttribute.Create(ClassFile, Offset)
  else
  begin
    Result := TAttribute.Create(ClassFile, Offset);
    Inc(Offset, Length);
  end;

  Check(Offset = Start + Length, ClassFile.FileName + ': Incorrect attribute size');
end;

constructor TAttribute.Create(ClassFile: TClassFile; Start: Integer);
var Offset: Integer;
begin
  inherited Create(ClassFile);
  FStart := Start;

  Offset := FStart - 6;
  FNameIndex := ClassFile.GetUInt16(Offset);
  FLength := ClassFile.GetUInt32(Offset);
end;

function TAttribute.GetName: TString;
begin
  Result := FClassFile.ConstantPoolUtf8[NameIndex - 1].Value
end;

// TConstantValueAttribute

constructor TConstantValueAttribute.Create(ClassFile: TClassFile; var Offset: Integer);
begin
  inherited Create(ClassFile, Offset);
  FValueIndex := ClassFile.GetUInt16(Offset);
end;

function TConstantValueAttribute.GetValue: TConstantPoolItem;
begin
  Result := FClassFile.ConstantPoolItem[FValueIndex - 1]
end;

// TSourceFileAttribute

constructor TSourceFileAttribute.Create(ClassFile: TClassFile; var Offset: Integer);
begin
  inherited Create(ClassFile, Offset);
  FSourceFileIndex := ClassFile.GetUInt16(Offset);
end;

function TSourceFileAttribute.GetSourceFile: TString;
begin
  Result := FClassFile.ConstantPoolUtf8[FSourceFileIndex - 1].Value
end;

// TCodeAttribute

constructor TCodeAttribute.Create(ClassFile: TClassFile; var Offset: Integer);
var ExceptionCount, AttributeCount, I: Integer;
begin
  inherited Create(ClassFile, Offset);

  FMaxStack := ClassFile.GetUInt16(Offset);
  FMaxLocals := ClassFile.GetUInt16(Offset);
  FCodeLength := ClassFile.GetUInt32(Offset);
  Inc(Offset, FCodeLength);

  ExceptionCount := ClassFile.GetUInt16(Offset);
  SetLength(FExceptions, ExceptionCount);
  for I := 0 to ExceptionCount - 1 do
  begin
    FExceptions[I].StartPc := ClassFile.GetUInt16(Offset);
    FExceptions[I].EndPc := ClassFile.GetUInt16(Offset);
    FExceptions[I].HandlerPc := ClassFile.GetUInt16(Offset);
    FExceptions[I].CatchType := ClassFile.GetUInt16(Offset);
  end;

  AttributeCount := ClassFile.GetUInt16(Offset);
  SetLength(FAttributes, AttributeCount);
  for I := 0 to AttributeCount - 1 do
    FAttributes[I] := TAttribute.CreateAttribute(ClassFile, Offset);
end;

function TCodeAttribute.GetCode(Index: Integer): Byte;
begin
  Check((Index >= 0) and (Index < Integer(FCodeLength)), ClassFile.FileName + ': Incorrect index of code');
  Result := ClassFile.FData[Start + 8 + Index];
end;

function TCodeAttribute.GetAttributeCount: Integer;
begin
  Result := System.Length(FAttributes);
end;

function TCodeAttribute.GetAttribute(Index: Integer): TAttribute;
begin
  Check((Index >= 0) and (Index < AttributeCount), ClassFile.FileName + ': Incorrect index of attribute');
  Result := FAttributes[Index];
end;

function TCodeAttribute.GetExceptionCount: Integer;
begin
  Result := System.Length(FExceptions);
end;

function TCodeAttribute.GetException(Index: Integer): TException;
begin
  Check((Index >= 0) and (Index < ExceptionCount), ClassFile.FileName + ': Incorrect index of exception');
  Result := FExceptions[Index];
end;

// TDeprecatedAttribute

constructor TDeprecatedAttribute.Create(ClassFile: TClassFile; var Offset: Integer);
begin
  inherited Create(ClassFile, Offset);
end;

// TSignatureAttribute

constructor TSignatureAttribute.Create(ClassFile: TClassFile; var Offset: Integer);
begin
  inherited Create(ClassFile, Offset);
  FDescriptorIndex := ClassFile.GetUInt16(Offset);
end;

function TSignatureAttribute.GetDescriptor: TConstantUtf8;
begin
  Result := ClassFile.ConstantPoolUtf8[DescriptorIndex - 1];
end;

// TInnerClassesAttribute

constructor TInnerClassesAttribute.Create(ClassFile: TClassFile; var Offset: Integer);
var Count, I: Integer;
begin
  inherited Create(ClassFile, Offset);

  Count := ClassFile.GetUInt16(Offset);
  SetLength(FClasses, Count);
  for I := 0 to Count - 1 do
  begin
    FClasses[I].InnerClassIndex := ClassFile.GetUInt16(Offset);
    FClasses[I].OuterClassIndex := ClassFile.GetUInt16(Offset);
    FClasses[I].InnerNameIndex := ClassFile.GetUInt16(Offset);
    FClasses[I].InnerClassAccessFlags := ClassFile.GetUInt16(Offset);
  end;
end;

function TInnerClassesAttribute.GetInnerClassCount: Integer;
begin
  Result := System.Length(FClasses);
end;

function TInnerClassesAttribute.GetInnerClass(Index: Integer): TInnerClass;
begin
  Check((Index >= 0) and (Index < InnerClassCount), ClassFile.FileName + ': Incorrect index of inner class');
  Result := FClasses[Index];
end;

function TInnerClassesAttribute.GetInnerName(Index: Integer): TString;
begin
  Check((Index >= 0) and (Index < InnerClassCount), ClassFile.FileName + ': Incorrect index of inner class');
  Result := ClassFile.ConstantPoolUtf8[FClasses[Index].InnerNameIndex - 1].Value;
end;

// TLineNumberTableAttribute

constructor TLineNumberTableAttribute.Create(ClassFile: TClassFile; var Offset: Integer);
var Count, I: Integer;
begin
  inherited Create(ClassFile, Offset);

  Count := ClassFile.GetUInt16(Offset);
  SetLength(FLineNumbers, Count);
  for I := 0 to Count - 1 do
  begin
    FLineNumbers[I].StartPc := ClassFile.GetUInt16(Offset);
    FLineNumbers[I].LineNumber := ClassFile.GetUInt16(Offset);
  end;
end;

function TLineNumberTableAttribute.GetLineNumberCount: Integer;
begin
  Result := System.Length(FLineNumbers);
end;

function TLineNumberTableAttribute.GetLineNumber(Index: Integer): TLineNumber;
begin
  Check((Index >= 0) and (Index < LineNumberCount), ClassFile.FileName + ': Incorrect index of line number');
  Result := FLineNumbers[Index];
end;

// TLocalVariableTableAttribute

constructor TLocalVariableTableAttribute.Create(ClassFile: TClassFile; var Offset: Integer);
var Count, I: Integer;
begin
  inherited Create(ClassFile, Offset);

  Count := ClassFile.GetUInt16(Offset);
  SetLength(FLocalVariables, Count);
  for I := 0 to Count - 1 do
  begin
    FLocalVariables[I].StartPc := ClassFile.GetUInt16(Offset);
    FLocalVariables[I].Length := ClassFile.GetUInt16(Offset);
    FLocalVariables[I].NameIndex := ClassFile.GetUInt16(Offset);
    FLocalVariables[I].DescriptorIndex := ClassFile.GetUInt16(Offset);
    FLocalVariables[I].Index := ClassFile.GetUInt16(Offset);
  end;
end;

function TLocalVariableTableAttribute.GetLocalVariableCount: Integer;
begin
  Result := System.Length(FLocalVariables);
end;

procedure TLocalVariableTableAttribute.CheckIndex(Index: Integer);
begin
  Check((Index >= 0) and (Index < LocalVariableCount), ClassFile.FileName + ': Incorrect index of local variable');
end;

function TLocalVariableTableAttribute.GetLocalVariable(Index: Integer): TLocalVariable;
begin
  CheckIndex(Index);
  Result := FLocalVariables[Index];
end;

function TLocalVariableTableAttribute.GetLocalVariableName(Index: Integer): TString;
begin
  CheckIndex(Index);
  Result := ClassFile.ConstantPoolUtf8[FLocalVariables[Index].NameIndex - 1].Value;
end;

function TLocalVariableTableAttribute.GetLocalVariableType(Index: Integer): TString;
begin
  CheckIndex(Index);
  Result := ClassFile.ConstantPoolUtf8[FLocalVariables[Index].DescriptorIndex - 1].Value;
end;

constructor TMethodParametersAttribute.Create(ClassFile: TClassFile; var Offset: Integer);
var Count, I: Integer;
begin
  inherited Create(ClassFile, Offset);

  Count := ClassFile.GetByte(Offset);
  SetLength(FMethodParameters, Count);
  for I := 0 to Count - 1 do
  begin
    FMethodParameters[I].NameIndex := ClassFile.GetUInt16(Offset);
    FMethodParameters[I].AccessFlags := ClassFile.GetUInt16(Offset);
  end;
end;

function TMethodParametersAttribute.GetMethodParameterCount: Integer;
begin
  Result := System.Length(FMethodParameters);
end;

procedure TMethodParametersAttribute.CheckIndex(Index: Integer);
begin
  Check((Index >= 0) and (Index < MethodParameterCount), ClassFile.FileName + ': Incorrect index of method parameter');
end;

function TMethodParametersAttribute.GetMethodParameter(Index: Integer): TMethodParameter;
begin
  CheckIndex(Index);
  Result := FMethodParameters[Index];
end;

function TMethodParametersAttribute.GetMethodParameterName(Index: Integer): TString;
var NameIndex: Integer;
begin
  CheckIndex(Index);
  NameIndex := FMethodParameters[Index].NameIndex;
  if NameIndex = 0 then
    Result := ''
  else
    Result := ClassFile.ConstantPoolUtf8[NameIndex - 1].Value;
end;

function TMethodParametersAttribute.GetIsMethodParameterFinal(Index: Integer): Boolean;
var AccessFlags: Word;
begin
  CheckIndex(Index);
  AccessFlags := FMethodParameters[Index].AccessFlags;
  Result := (AccessFlags and JVM_ACC_FINAL) <> 0
end;

function TMethodParametersAttribute.GetIsMethodParameterSynthetic(Index: Integer): Boolean;
var AccessFlags: Word;
begin
  CheckIndex(Index);
  AccessFlags := FMethodParameters[Index].AccessFlags;
  Result := (AccessFlags and JVM_ACC_SYNTHETIC) <> 0
end;

function TMethodParametersAttribute.GetIsMethodParameterMandated(Index: Integer): Boolean;
const JVM_ACC_MANDATED = $8000;
var AccessFlags: Word;
begin
  CheckIndex(Index);
  AccessFlags := FMethodParameters[Index].AccessFlags;
  Result := (AccessFlags and JVM_ACC_MANDATED) <> 0
end;

end.