// ---------------------------------------------------------------------
//
// JavaBridge
//
// Copyright (c) 2012-2021 WINSOFT
//
// ---------------------------------------------------------------------

//{$DEFINE TRIAL} // trial version, comment this line for full version

unit JavaBridge;

{$ifdef CONDITIONALEXPRESSIONS}
  {$if CompilerVersion >= 14}
    {$define D6PLUS} // Delphi 6 or higher
  {$ifend}

  {$if CompilerVersion >= 20}
    {$define D2009PLUS} // Delphi 2009 or higher
  {$ifend}
{$endif}

{$ifndef CONDITIONALEXPRESSIONS}
  {$define MSWINDOWS}
{$endif CONDITIONALEXPRESSIONS}

{$ifdef FPC}
  {$define MSWINDOWS}
{$endif FPC}

interface

uses JNI {$ifdef MSWINDOWS}, Windows {$endif MSWINDOWS}, SysUtils;

type
  WString = {$ifdef D2009PLUS} string {$else} WideString {$endif D2009PLUS};

  EJavaBridgeError = class(Exception);

  TJValueDynArray = array of JValue;

  TJavaBridge = class
  private
    FEnvironment: PJNIEnv;
    FJNI_GetDefaultJavaVMInitArgs: JNI.JNI_GetDefaultJavaVMInitArgs;
    FJNI_CreateJavaVM: JNI.JNI_CreateJavaVM;
    FJNI_GetCreatedJavaVMs: JNI.JNI_GetCreatedJavaVMs;
    FJavaVm: PJavaVM;
    FJavaVmArguments: JavaVMInitArgs;
    FLibraryModule: HMODULE;
    FLibraryName: string;

    procedure Check(Value: Boolean; const Message: string);
    function CheckGetProcAddress(const Name: string): Pointer;
    procedure CreateJavaVm;
    function GetJavaVmCreated: Boolean;
    function GetLibraryLoaded: Boolean;
    procedure LoadLibrary;
    procedure ReleaseJavaVm;
    procedure SetJavaVmCreated(Value: Boolean);
    procedure SetLibraryLoaded(Value: Boolean);
    procedure UnloadLibrary;
  public
    constructor Create; overload;
    constructor Create(Environment: PJNIEnv); overload;
    destructor Destroy; override;

    function CreateJArguments(const Arguments: array of const): TJValueDynArray;
    function CreateJString(const Text: AnsiString): JString; overload;
    {$ifdef D6PLUS}
    function CreateJString(const Text: WString): JString; overload;
    {$else}
    function CreateJStringW(const Text: WString): JString;
    {$endif D6PLUS}
    procedure DefaultJavaVmArguments;

    function GetVersion: JInt;
    function DefineClass(const Name: PAnsiChar; Loader: JObject; const Buf: PJByte; Len: JSize): JClass;
    function FindClass(const Name: PAnsiChar): JClass;
    function FromReflectedMethod(Method: JObject): JMethodID;
    function FromReflectedField(Field: JObject): JFieldID;
    function ToReflectedMethod(Cls: JClass; MethodID: JMethodID; IsStatic: JBoolean): JObject;
    function GetSuperclass(Sub: JClass): JClass;
    function IsAssignableFrom(Sub, Sup: JClass): JBoolean;
    function ToReflectedField(Cls: JClass; FieldID: JFieldID; IsStatic: JBoolean): JObject;
    function Throw(Obj: JThrowable): JInt;
    function ThrowNew(Clazz: JClass; const Msg: PAnsiChar): JInt;
    function ExceptionOccurred: JThrowable;
    procedure ExceptionDescribe;
    procedure ExceptionClear;
    procedure FatalError(const Msg: PAnsiChar);
    function PushLocalFrame(Capacity: JInt): JInt;
    function PopLocalFrame(Result_: JObject): JObject;
    function NewGlobalRef(LObj: JObject): JObject;
    procedure DeleteGlobalRef(GRef: JObject);
    procedure DeleteLocalRef(Obj: JObject);
    function IsSameObject(Obj1, Obj2: JObject): JBoolean;
    function NewLocalRef(Ref: JObject): JObject;
    function EnsureLocalCapacity(Capacity: JInt): JInt;
    function AllocObject(Clazz: JClass): JObject;
    function NewObject(Clazz: JClass; MethodID: JMethodID { ... } ): JObject;
    function NewObjectV(Clazz: JClass; MethodID: JMethodID; Args: VaList): JObject;
    function NewObjectA(Clazz: JClass; MethodID: JMethodID; const Args: PJValue): JObject;
    function GetObjectClass(Obj: JObject): JClass;
    function IsInstanceOf(Obj: JObject; Clazz: JClass): JBoolean;
    function GetMethodID(Clazz: JClass; const Name, Sig: PAnsiChar): JMethodID;
    function CallObjectMethod(Obj: JObject; MethodID: JMethodID { ... } ): JObject;
    function CallObjectMethodV(Obj: JObject; MethodID: JMethodID; Args: VaList): JObject;
    function CallObjectMethodA(Obj: JObject; MethodID: JMethodID; const Args: PJValue): JObject;
    function CallBooleanMethod(Obj: JObject; MethodID: JMethodID { ... } ): JBoolean;
    function CallBooleanMethodV(Obj: JObject; MethodID: JMethodID; Args: VaList): JBoolean;
    function CallBooleanMethodA(Obj: JObject; MethodID: JMethodID; const Args: PJValue): JBoolean;
    function CallByteMethod(Obj: JObject; MethodID: JMethodID { ... } ): JByte;
    function CallByteMethodV(Obj: JObject; MethodID: JMethodID; Args: VaList): JByte;
    function CallByteMethodA(Obj: JObject; MethodID: JMethodID; const Args: PJValue): JByte;
    function CallCharMethod(Obj: JObject; MethodID: JMethodID { ... } ): JChar;
    function CallCharMethodV(Obj: JObject; MethodID: JMethodID; Args: VaList): JChar;
    function CallCharMethodA(Obj: JObject; MethodID: JMethodID; const Args: PJValue): JChar;
    function CallShortMethod(Obj: JObject; MethodID: JMethodID { ... } ): JShort;
    function CallShortMethodV(Obj: JObject; MethodID: JMethodID; Args: VaList): JShort;
    function CallShortMethodA(Obj: JObject; MethodID: JMethodID; const Args: PJValue): JShort;
    function CallIntMethod(Obj: JObject; MethodID: JMethodID { ... } ): JInt;
    function CallIntMethodV(Obj: JObject; MethodID: JMethodID; Args: VaList): JInt;
    function CallIntMethodA(Obj: JObject; MethodID: JMethodID; const Args: PJValue): JInt;
    function CallLongMethod(Obj: JObject; MethodID: JMethodID { ... } ): JLong;
    function CallLongMethodV(Obj: JObject; MethodID: JMethodID; Args: VaList): JLong;
    function CallLongMethodA(Obj: JObject; MethodID: JMethodID; const Args: PJValue): JLong;
    function CallFloatMethod(Obj: JObject; MethodID: JMethodID { ... } ): JFloat;
    function CallFloatMethodV(Obj: JObject; MethodID: JMethodID; Args: VaList): JFloat;
    function CallFloatMethodA(Obj: JObject; MethodID: JMethodID; const Args: PJValue): JFloat;
    function CallDoubleMethod(Obj: JObject; MethodID: JMethodID { ... } ): JDouble;
    function CallDoubleMethodV(Obj: JObject; MethodID: JMethodID; Args: VaList): JDouble;
    function CallDoubleMethodA(Obj: JObject; MethodID: JMethodID; const Args: PJValue): JDouble;
    procedure CallVoidMethod(Obj: JObject; MethodID: JMethodID { ... } );
    procedure CallVoidMethodV(Obj: JObject; MethodID: JMethodID; Args: VaList);
    procedure CallVoidMethodA(Obj: JObject; MethodID: JMethodID; const Args: PJValue);
    function CallNonvirtualObjectMethod(Obj: JObject; Clazz: JClass; MethodID: JMethodID { ... } ): JObject;
    function CallNonvirtualObjectMethodV(Obj: JObject; Clazz: JClass; MethodID: JMethodID; Args: VaList): JObject;
    function CallNonvirtualObjectMethodA(Obj: JObject; Clazz: JClass; MethodID: JMethodID; const Args: PJValue): JObject;
    function CallNonvirtualBooleanMethod(Obj: JObject; Clazz: JClass; MethodID: JMethodID { ... } ): JBoolean;
    function CallNonvirtualBooleanMethodV(Obj: JObject; Clazz: JClass; MethodID: JMethodID; Args: VaList): JBoolean;
    function CallNonvirtualBooleanMethodA(Obj: JObject; Clazz: JClass; MethodID: JMethodID; const Args: PJValue): JBoolean;
    function CallNonvirtualByteMethod(Obj: JObject; Clazz: JClass; MethodID: JMethodID { ... } ): JByte;
    function CallNonvirtualByteMethodV(Obj: JObject; Clazz: JClass; MethodID: JMethodID; Args: VaList): JByte;
    function CallNonvirtualByteMethodA(Obj: JObject; Clazz: JClass; MethodID: JMethodID; const Args: PJValue): JByte;
    function CallNonvirtualCharMethod(Obj: JObject; Clazz: JClass; MethodID: JMethodID { ... } ): JChar;
    function CallNonvirtualCharMethodV(Obj: JObject; Clazz: JClass; MethodID: JMethodID; Args: VaList): JChar;
    function CallNonvirtualCharMethodA(Obj: JObject; Clazz: JClass; MethodID: JMethodID; const Args: PJValue): JChar;
    function CallNonvirtualShortMethod(Obj: JObject; Clazz: JClass; MethodID: JMethodID { ... } ): JShort;
    function CallNonvirtualShortMethodV(Obj: JObject; Clazz: JClass; MethodID: JMethodID; Args: VaList): JShort;
    function CallNonvirtualShortMethodA(Obj: JObject; Clazz: JClass; MethodID: JMethodID; const Args: PJValue): JShort;
    function CallNonvirtualIntMethod(Obj: JObject; Clazz: JClass; MethodID: JMethodID { ... } ): JInt;
    function CallNonvirtualIntMethodV(Obj: JObject; Clazz: JClass; MethodID: JMethodID; Args: VaList): JInt;
    function CallNonvirtualIntMethodA(Obj: JObject; Clazz: JClass; MethodID: JMethodID; const Args: PJValue): JInt;
    function CallNonvirtualLongMethod(Obj: JObject; Clazz: JClass; MethodID: JMethodID { ... } ): JLong;
    function CallNonvirtualLongMethodV(Obj: JObject; Clazz: JClass; MethodID: JMethodID; Args: VaList): JLong;
    function CallNonvirtualLongMethodA(Obj: JObject; Clazz: JClass; MethodID: JMethodID; const Args: PJValue): JLong;
    function CallNonvirtualFloatMethod(Obj: JObject; Clazz: JClass; MethodID: JMethodID { ... } ): JFloat;
    function CallNonvirtualFloatMethodV(Obj: JObject; Clazz: JClass; MethodID: JMethodID; Args: VaList): JFloat;
    function CallNonvirtualFloatMethodA(Obj: JObject; Clazz: JClass; MethodID: JMethodID; const Args: PJValue): JFloat;
    function CallNonvirtualDoubleMethod(Obj: JObject; Clazz: JClass; MethodID: JMethodID { ... } ): JDouble;
    function CallNonvirtualDoubleMethodV(Obj: JObject; Clazz: JClass; MethodID: JMethodID; Args: VaList): JDouble;
    function CallNonvirtualDoubleMethodA(Obj: JObject; Clazz: JClass; MethodID: JMethodID; const Args: PJValue): JDouble;
    procedure CallNonvirtualVoidMethod(Obj: JObject; Clazz: JClass; MethodID: JMethodID { ... } );
    procedure CallNonvirtualVoidMethodV(Obj: JObject; Clazz: JClass; MethodID: JMethodID; Args: VaList);
    procedure CallNonvirtualVoidMethodA(Obj: JObject; Clazz: JClass; MethodID: JMethodID; const Args: PJValue);
    function GetFieldID(Clazz: JClass; const Name, Sig: PAnsiChar): JFieldID;
    function GetObjectField(Obj: JObject; FieldID: JFieldID): JObject;
    function GetBooleanField(Obj: JObject; FieldID: JFieldID): JBoolean;
    function GetByteField(Obj: JObject; FieldID: JFieldID): JByte;
    function GetCharField(Obj: JObject; FieldID: JFieldID): JChar;
    function GetShortField(Obj: JObject; FieldID: JFieldID): JShort;
    function GetIntField(Obj: JObject; FieldID: JFieldID): JInt;
    function GetLongField(Obj: JObject; FieldID: JFieldID): JLong;
    function GetFloatField(Obj: JObject; FieldID: JFieldID): JFloat;
    function GetDoubleField(Obj: JObject; FieldID: JFieldID): JDouble;
    procedure SetObjectField(Obj: JObject; FieldID: JFieldID; Val: JObject);
    procedure SetBooleanField(Obj: JObject; FieldID: JFieldID; Val: JBoolean);
    procedure SetByteField(Obj: JObject; FieldID: JFieldID; Val: JByte);
    procedure SetCharField(Obj: JObject; FieldID: JFieldID; Val: JChar);
    procedure SetShortField(Obj: JObject; FieldID: JFieldID; Val: JShort);
    procedure SetIntField(Obj: JObject; FieldID: JFieldID; Val: JInt);
    procedure SetLongField(Obj: JObject; FieldID: JFieldID; Val: JLong);
    procedure SetFloatField(Obj: JObject; FieldID: JFieldID; Val: JFloat);
    procedure SetDoubleField(Obj: JObject; FieldID: JFieldID; Val: JDouble);
    function GetStaticMethodID(Clazz: JClass; const Name, Sig: PAnsiChar): JMethodID;
    function CallStaticObjectMethod(Clazz: JClass; MethodID: JMethodID { ... } ): JObject;
    function CallStaticObjectMethodV(Clazz: JClass; MethodID: JMethodID; Args: VaList): JObject;
    function CallStaticObjectMethodA(Clazz: JClass; MethodID: JMethodID; const Args: PJValue): JObject;
    function CallStaticBooleanMethod(Clazz: JClass; MethodID: JMethodID { ... } ): JBoolean;
    function CallStaticBooleanMethodV(Clazz: JClass; MethodID: JMethodID; Args: VaList): JBoolean;
    function CallStaticBooleanMethodA(Clazz: JClass; MethodID: JMethodID; const Args: PJValue): JBoolean;
    function CallStaticByteMethod(Clazz: JClass; MethodID: JMethodID { ... } ): JByte;
    function CallStaticByteMethodV(Clazz: JClass; MethodID: JMethodID; Args: VaList): JByte;
    function CallStaticByteMethodA(Clazz: JClass; MethodID: JMethodID; const Args: PJValue): JByte;
    function CallStaticCharMethod(Clazz: JClass; MethodID: JMethodID { ... } ): JChar;
    function CallStaticCharMethodV(Clazz: JClass; MethodID: JMethodID; Args: VaList): JChar;
    function CallStaticCharMethodA(Clazz: JClass; MethodID: JMethodID; const Args: PJValue): JChar;
    function CallStaticShortMethod(Clazz: JClass; MethodID: JMethodID { ... } ): JShort;
    function CallStaticShortMethodV(Clazz: JClass; MethodID: JMethodID; Args: VaList): JShort;
    function CallStaticShortMethodA(Clazz: JClass; MethodID: JMethodID; const Args: PJValue): JShort;
    function CallStaticIntMethod(Clazz: JClass; MethodID: JMethodID { ... } ): JInt;
    function CallStaticIntMethodV(Clazz: JClass; MethodID: JMethodID; Args: VaList): JInt;
    function CallStaticIntMethodA(Clazz: JClass; MethodID: JMethodID; const Args: PJValue): JInt;
    function CallStaticLongMethod(Clazz: JClass; MethodID: JMethodID { ... } ): JLong;
    function CallStaticLongMethodV(Clazz: JClass; MethodID: JMethodID; Args: VaList): JLong;
    function CallStaticLongMethodA(Clazz: JClass; MethodID: JMethodID; const Args: PJValue): JLong;
    function CallStaticFloatMethod(Clazz: JClass; MethodID: JMethodID { ... } ): JFloat;
    function CallStaticFloatMethodV(Clazz: JClass; MethodID: JMethodID; Args: VaList): JFloat;
    function CallStaticFloatMethodA(Clazz: JClass; MethodID: JMethodID; const Args: PJValue): JFloat;
    function CallStaticDoubleMethod(Clazz: JClass; MethodID: JMethodID { ... } ): JDouble;
    function CallStaticDoubleMethodV(Clazz: JClass; MethodID: JMethodID; Args: VaList): JDouble;
    function CallStaticDoubleMethodA(Clazz: JClass; MethodID: JMethodID; const Args: PJValue): JDouble;
    procedure CallStaticVoidMethod(Cls: JClass; MethodID: JMethodID { ... } );
    procedure CallStaticVoidMethodV(Cls: JClass; MethodID: JMethodID; Args: VaList);
    procedure CallStaticVoidMethodA(Cls: JClass; MethodID: JMethodID; const Args: PJValue);
    function GetStaticFieldID(Clazz: JClass; const Name, Sig: PAnsiChar): JFieldID;
    function GetStaticObjectField(Clazz: JClass; FieldID: JFieldID): JObject;
    function GetStaticBooleanField(Clazz: JClass; FieldID: JFieldID): JBoolean;
    function GetStaticByteField(Clazz: JClass; FieldID: JFieldID): JByte;
    function GetStaticCharField(Clazz: JClass; FieldID: JFieldID): JChar;
    function GetStaticShortField(Clazz: JClass; FieldID: JFieldID): JShort;
    function GetStaticIntField(Clazz: JClass; FieldID: JFieldID): JInt;
    function GetStaticLongField(Clazz: JClass; FieldID: JFieldID): JLong;
    function GetStaticFloatField(Clazz: JClass; FieldID: JFieldID): JFloat;
    function GetStaticDoubleField(Clazz: JClass; FieldID: JFieldID): JDouble;
    procedure SetStaticObjectField(Clazz: JClass; FieldID: JFieldID; Value: JObject);
    procedure SetStaticBooleanField(Clazz: JClass; FieldID: JFieldID; Value: JBoolean);
    procedure SetStaticByteField(Clazz: JClass; FieldID: JFieldID; Value: JByte);
    procedure SetStaticCharField(Clazz: JClass; FieldID: JFieldID; Value: JChar);
    procedure SetStaticShortField(Clazz: JClass; FieldID: JFieldID; Value: JShort);
    procedure SetStaticIntField(Clazz: JClass; FieldID: JFieldID; Value: JInt);
    procedure SetStaticLongField(Clazz: JClass; FieldID: JFieldID; Value: JLong);
    procedure SetStaticFloatField(Clazz: JClass; FieldID: JFieldID; Value: JFloat);
    procedure SetStaticDoubleField(Clazz: JClass; FieldID: JFieldID; Value: JDouble);
    function NewString(const Unicode: PJChar; Len: JSize): JString;
    function GetStringLength(Str: JString): JSize;
    function GetStringChars(Str: JString; var IsCopy: JBoolean): PJChar;
    procedure ReleaseStringChars(Str: JString; const Chars: PJChar);
    function NewStringUTF(const Utf: PAnsiChar): JString;
    function GetStringUTFLength(Str: JString): JSize;
    function GetStringUTFChars(Str: JString; var IsCopy: JBoolean): PAnsiChar;
    procedure ReleaseStringUTFChars(Str: JString; const Chars: PAnsiChar);
    function GetArrayLength(Array_: JArray): JSize;
    function NewObjectArray(Len: JSize; Clazz: JClass; Init: JObject): JObjectArray;
    function GetObjectArrayElement(Array_: JObjectArray; Index: JSize): JObject;
    procedure SetObjectArrayElement(Array_: JObjectArray; Index: JSize; Val: JObject);
    function NewBooleanArray(Len: JSize): JBooleanArray;
    function NewByteArray(Len: JSize): JByteArray;
    function NewCharArray(Len: JSize): JCharArray;
    function NewShortArray(Len: JSize): JShortArray;
    function NewIntArray(Len: JSize): JIntArray;
    function NewLongArray(Len: JSize): JLongArray;
    function NewFloatArray(Len: JSize): JFloatArray;
    function NewDoubleArray(Len: JSize): JDoubleArray;
    function GetBooleanArrayElements(Array_: JBooleanArray; var IsCopy: JBoolean): PJBoolean;
    function GetByteArrayElements(Array_: JByteArray; var IsCopy: JBoolean): PJByte;
    function GetCharArrayElements(Array_: JCharArray; var IsCopy: JBoolean): PJChar;
    function GetShortArrayElements(Array_: JShortArray; var IsCopy: JBoolean): PJShort;
    function GetIntArrayElements(Array_: JIntArray; var IsCopy: JBoolean): PJInt;
    function GetLongArrayElements(Array_: JLongArray; var IsCopy: JBoolean): PJLong;
    function GetFloatArrayElements(Array_: JFloatArray; var IsCopy: JBoolean): PJFloat;
    function GetDoubleArrayElements(Array_: JDoubleArray; var IsCopy: JBoolean): PJDouble;
    procedure ReleaseBooleanArrayElements(Array_: JBooleanArray; Elems: PJBoolean; Mode: JInt);
    procedure ReleaseByteArrayElements(Array_: JByteArray; Elems: PJByte; Mode: JInt);
    procedure ReleaseCharArrayElements(Array_: JCharArray; Elems: PJChar; Mode: JInt);
    procedure ReleaseShortArrayElements(Array_: JShortArray; Elems: PJShort; Mode: JInt);
    procedure ReleaseIntArrayElements(Array_: JIntArray; Elems: PJInt; Mode: JInt);
    procedure ReleaseLongArrayElements(Array_: JLongArray; Elems: PJLong; Mode: JInt);
    procedure ReleaseFloatArrayElements(Array_: JFloatArray; Elems: PJFloat; Mode: JInt);
    procedure ReleaseDoubleArrayElements(Array_: JDoubleArray; Elems: PJDouble; Mode: JInt);
    procedure GetBooleanArrayRegion(Array_: JBooleanArray; Start, Len: JSize; Buf: PJBoolean);
    procedure GetByteArrayRegion(Array_: JByteArray; Start, Len: JSize; Buf: PJByte);
    procedure GetCharArrayRegion(Array_: JCharArray; Start, Len: JSize; Buf: PJChar);
    procedure GetShortArrayRegion(Array_: JShortArray; Start, Len: JSize; Buf: PJShort);
    procedure GetIntArrayRegion(Array_: JIntArray; Start, Len: JSize; Buf: PJInt);
    procedure GetLongArrayRegion(Array_: JLongArray; Start, Len: JSize; Buf: PJLong);
    procedure GetFloatArrayRegion(Array_: JFloatArray; Start, Len: JSize; Buf: PJFloat);
    procedure GetDoubleArrayRegion(Array_: JDoubleArray; Start, Len: JSize; Buf: PJDouble);
    procedure SetBooleanArrayRegion(Array_: JBooleanArray; Start, Len: JSize; const Buf: PJBoolean);
    procedure SetByteArrayRegion(Array_: JByteArray; Start, Len: JSize; const Buf: PJByte);
    procedure SetCharArrayRegion(Array_: JCharArray; Start, Len: JSize; const Buf: PJChar);
    procedure SetShortArrayRegion(Array_: JShortArray; Start, Len: JSize; const Buf: PJShort);
    procedure SetIntArrayRegion(Array_: JIntArray; Start, Len: JSize; const Buf: PJInt);
    procedure SetLongArrayRegion(Array_: JLongArray; Start, Len: JSize; const Buf: PJLong);
    procedure SetFloatArrayRegion(Array_: JFloatArray; Start, Len: JSize; const Buf: PJFloat);
    procedure SetDoubleArrayRegion(Array_: JDoubleArray; Start, Len: JSize; const Buf: PJDouble);
    function RegisterNatives(Clazz: JClass; const Methods: PJNINativeMethod; NMethods: JInt): JInt;
    function UnregisterNatives(Clazz: JClass): JInt;
    function MonitorEnter(Obj: JObject): JInt;
    function MonitorExit(Obj: JObject): JInt;
    function GetJavaVM(out VM: PJavaVM): JInt;
    procedure GetStringRegion(Str: JString; Start, Len: JSize; Buf: PJChar);
    procedure GetStringUTFRegion(Str: JString; Start, Len: JSize; Buf: PAnsiChar);
    function GetPrimitiveArrayCritical(Array_: JArray; var IsCopy: JBoolean): Pointer;
    procedure ReleasePrimitiveArrayCritical(Array_: JArray; Carray: Pointer; Mode: JInt);
    function GetStringCritical(String_: JString; var IsCopy: JBoolean): PJChar;
    procedure ReleaseStringCritical(String_: JString; const CString: PJChar);
    function NewWeakGlobalRef(Obj: JObject): JWeak;
    procedure DeleteWeakGlobalRef(Ref: JWeak);
    function ExceptionCheck: JBoolean;
    function NewDirectByteBuffer(Address: Pointer; Capacity: JLong): JObject;
    function GetDirectBufferAddress(Buf: JObject): Pointer;
    function GetDirectBufferCapacity(Buf: JObject): JLong;
    function GetObjectRefType(Obj: JObject): JObjectRefType;
    function GetModule(Clazz: JClass): JObject;

    property Environment: PJNIEnv read FEnvironment;
    property JavaVm: PJavaVM read FJavaVm;
    property JavaVmArguments: JavaVMInitArgs read FJavaVmArguments;
    property JavaVmCreated: Boolean read GetJavaVmCreated write SetJavaVmCreated;
    property LibraryLoaded: Boolean read GetLibraryLoaded write SetLibraryLoaded;
    property LibraryModule: HMODULE read FLibraryModule;
    property LibraryName: string read FLibraryName write FLibraryName;
  end;

implementation

{$ifdef MSWINDOWS}
uses Registry {$ifdef TRIAL} {$ifndef FPC}, Forms {$endif FPC} {$endif TRIAL};
{$endif MSWINDOWS}

{$ifdef TRIAL}
var WasTrial: Boolean;

procedure ShowTrial;
begin
  if not WasTrial then
  begin
    WasTrial := True;

{$ifdef MSWINDOWS}
    MessageBox({$ifdef FPC} 0 {$else} Application.Handle {$endif FPC},
      'A trial version of JavaBridge started.' + #13#13 +
      'Please note that trial version is supposed to be used for evaluation only. ' +
      'If you wish to distribute JavaBridge as part of your application, ' +
      'you must register from website at https://www.winsoft.sk.' + #13#13 +
      'Thank you for trialing JavaBridge.',
      'JavaBridge 3.0, Copyright (c) 2012-2021 WINSOFT',
       MB_OK or MB_ICONINFORMATION);
{$else}
    WriteLn('JavaBridge 3.0, Copyright (c) 2012-2021 WINSOFT');
    WriteLn;
    WriteLn('A trial version of JavaBridge started.');
    WriteLn;
    WriteLn('Please note that trial version is supposed to be used for evaluation only.');
    WriteLn('If you wish to distribute JavaBridge as part of your application,');
    WriteLn('you must register from website at https://www.winsoft.sk.');
    WriteLn;
    WriteLn('Thank you for trialing JavaBridge.');
    WriteLn;
    WriteLn('Press Enter key to continue');
    ReadLn;
{$endif MSWINDOWS}
  end;
end;
{$endif TRIAL}

function GetLibraryName: string;
{$ifdef MSWINDOWS}
var
  Registry: TRegistry;
  Version: string;
{$endif MSWINDOWS}
begin
{$ifdef MSWINDOWS}
  Result := 'jvm.dll';

  Registry := TRegistry.Create(KEY_READ);
  try
    Registry.RootKey := HKEY_LOCAL_MACHINE;
    if Registry.OpenKey('\SOFTWARE\JavaSoft\Java Runtime Environment', False) then
    begin
      Version := Registry.ReadString('CurrentVersion');
      if Version <> '' then
        if Registry.OpenKey(Version, False) then
          Result := Registry.ReadString('RuntimeLib');
    end
    else if Registry.OpenKey('\SOFTWARE\JavaSoft\JDK', False) then
    begin
      Version := Registry.ReadString('CurrentVersion');
      if Version <> '' then
        if Registry.OpenKey(Version, False) then
          Result := Registry.ReadString('JavaHome') + '\bin\server\jvm.dll';
    end
  finally
    Registry.Free;
  end;
{$endif MSWINDOWS}

{$ifdef MACOS64}
  Result := '/Library/Java/JavaVirtualMachines/jdk-17.0.1.jdk/Contents/MacOS/libjli.dylib';
{$endif MACOS64}

{$ifdef LINUX64}
  Result := '/snap/openjdk/current/jdk/lib/server/libjvm.so'; //'/snap/openjdk/current/jdk/lib/libjli.so';
{$endif LINUX64}
end;

constructor TJavaBridge.Create;
begin
  inherited Create;
  FJavaVmArguments.version := JNI_VERSION_1_8;

  {$ifdef TRIAL}
    ShowTrial;
  {$endif TRIAL}
end;

constructor TJavaBridge.Create(Environment: PJNIEnv);
begin
  inherited Create;
  FEnvironment := Environment;
  Check(GetJavaVM(FJavaVM) = JNI_OK, 'Cannot retrieve JavaVM');

  {$ifdef TRIAL}
    ShowTrial;
  {$endif TRIAL}
end;

destructor TJavaBridge.Destroy;
begin
  LibraryLoaded := False;
  inherited Destroy;
end;

procedure TJavaBridge.Check(Value: Boolean; const Message: string);
begin
  if not Value then
    raise EJavaBridgeError.Create(Message);
end;

function TJavaBridge.CheckGetProcAddress(const Name: string): Pointer;
var ErrorMessage: string;
begin
  Result := GetProcAddress(FLibraryModule, PChar(Name));
  if Result = nil then
  begin
    ErrorMessage := SysErrorMessage(GetLastError) + ': ' + Name;
    UnloadLibrary;
    Check(False, ErrorMessage);
  end;
end;

function TJavaBridge.GetLibraryLoaded: Boolean;
begin
  Result := FLibraryModule <> 0;
end;

procedure TJavaBridge.SetLibraryLoaded(Value: Boolean);
begin
  if Value then
    LoadLibrary
  else
    UnloadLibrary;
end;

procedure TJavaBridge.LoadLibrary;
begin
  if not LibraryLoaded then
  begin
    {$ifdef TRIAL}
      ShowTrial;
    {$endif TRIAL}

    if FLibraryName = '' then
      FLibraryName := GetLibraryName;
    FLibraryModule := SafeLoadLibrary(PChar(FLibraryName));
    Check(FLibraryModule <> 0, SysErrorMessage(GetLastError) + ': ' + FLibraryName);
    FJNI_GetDefaultJavaVMInitArgs := CheckGetProcAddress('JNI_GetDefaultJavaVMInitArgs');
    FJNI_CreateJavaVM := CheckGetProcAddress('JNI_CreateJavaVM');
    FJNI_GetCreatedJavaVMs := CheckGetProcAddress('JNI_GetCreatedJavaVMs');
  end;
end;

procedure TJavaBridge.UnloadLibrary;
begin
  if LibraryLoaded then
  begin
    JavaVmCreated := False;
    FreeLibrary(FLibraryModule); // ignore result value
    FLibraryModule := 0;
  end;
end;

function TJavaBridge.GetJavaVmCreated: Boolean;
begin
  Result := (FJavaVM <> nil) or (FEnvironment <> nil); // JavaVM can be initialized using Environment
end;

procedure TJavaBridge.SetJavaVmCreated(Value: Boolean);
begin
  if Value then
    CreateJavaVm
  else
    ReleaseJavaVm;
end;

procedure TJavaBridge.CreateJavaVm;
begin
  if not JavaVmCreated then
  begin
    {$ifdef TRIAL}
      ShowTrial;
    {$endif TRIAL}

    LibraryLoaded := True;
    Check(FJNI_CreateJavaVM(FJavaVM, FEnvironment, FJavaVmArguments) = JNI_OK, 'Cannot create JavaVM');
  end;
end;

procedure TJavaBridge.ReleaseJavaVm;
begin
  if LibraryLoaded and JavaVmCreated then // JVM can't be released when initialized using Environment
  begin
    FJavaVm^.DestroyJavaVM(FJavaVm); // ignore result value
    FJavaVm := nil;
    FEnvironment := nil;
  end;
end;

function TJavaBridge.CreateJString(const Text: AnsiString): JString;
begin
  Result := NewStringUTF(PAnsiChar(Text));
end;

{$ifdef D6PLUS}
function TJavaBridge.CreateJString(const Text: WString): JString;
{$else}
function TJavaBridge.CreateJStringW(const Text: WString): JString;
{$endif D6PLUS}
begin
  Result := NewString(PJChar(PWideChar(Text)), Length(Text));
end;

function TJavaBridge.CreateJArguments(const Arguments: array of const): TJValueDynArray;
var I: Integer;
begin
  SetLength(Result, Length(Arguments));
  for I := Low(Arguments) to High(Arguments) do
    case Arguments[I].VType of
      vtInteger:
        Result[I].I := Arguments[I].VInteger;
      vtBoolean:
        Result[I].Z := Arguments[I].VBoolean;
      vtChar:
        Result[I].C := WideChar(Arguments[I].VChar);
      vtExtended:
        Result[I].D := Arguments[I].VExtended^;
      vtPointer:
        Result[I].L := JObject(Arguments[I].VPointer);
      vtWideChar:
        Result[I].C := Arguments[I].VWideChar;
      vtAnsiString:
        Result[I].L := CreateJString(PAnsiChar(Arguments[I].VAnsiString));
      vtWideString:
        Result[I].L := CreateJString(PWideChar(Arguments[I].VWideString));
      vtInt64:
        Result[I].J := Arguments[I].VInt64^;
    else
      Check(False, 'Unsupported argument type');
    end;
end;

procedure TJavaBridge.DefaultJavaVmArguments;
begin
  LibraryLoaded := True;
  Check(FJNI_GetDefaultJavaVMInitArgs(FJavaVmArguments) = JNI_OK, 'Cannot retrieve default arguments');
end;

function TJavaBridge.GetVersion: JInt;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetVersion(FEnvironment);
end;

function TJavaBridge.DefineClass(const Name: PAnsiChar; Loader: JObject; const Buf: PJByte; Len: JSize): JClass;
begin
  CreateJavaVM;
  Result := FEnvironment^.DefineClass(FEnvironment, Name, Loader, Buf, Len);
end;

function TJavaBridge.FindClass(const Name: PAnsiChar): JClass;
begin
  CreateJavaVM;
  Result := FEnvironment^.FindClass(FEnvironment, Name);
end;

function TJavaBridge.FromReflectedMethod(Method: JObject): JMethodID;
begin
  CreateJavaVM;
  Result := FEnvironment^.FromReflectedMethod(FEnvironment, Method);
end;

function TJavaBridge.FromReflectedField(Field: JObject): JFieldID;
begin
  CreateJavaVM;
  Result := FEnvironment^.FromReflectedField(FEnvironment, Field);
end;

function TJavaBridge.ToReflectedMethod(Cls: JClass; MethodID: JMethodID; IsStatic: JBoolean): JObject;
begin
  CreateJavaVM;
  Result := FEnvironment^.ToReflectedMethod(FEnvironment, Cls, MethodID, IsStatic);
end;

function TJavaBridge.GetSuperclass(Sub: JClass): JClass;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetSuperclass(FEnvironment, Sub);
end;

function TJavaBridge.IsAssignableFrom(Sub, Sup: JClass): JBoolean;
begin
  CreateJavaVM;
  Result := FEnvironment^.IsAssignableFrom(FEnvironment, Sub, Sup);
end;

function TJavaBridge.ToReflectedField(Cls: JClass; FieldID: JFieldID; IsStatic: JBoolean): JObject;
begin
  CreateJavaVM;
  Result := FEnvironment^.ToReflectedField(FEnvironment, Cls, FieldID, IsStatic);
end;

function TJavaBridge.Throw(Obj: JThrowable): JInt;
begin
  CreateJavaVM;
  Result := FEnvironment^.Throw(FEnvironment, Obj);
end;

function TJavaBridge.ThrowNew(Clazz: JClass; const Msg: PAnsiChar): JInt;
begin
  CreateJavaVM;
  Result := FEnvironment^.ThrowNew(FEnvironment, Clazz, Msg);
end;

function TJavaBridge.ExceptionOccurred: JThrowable;
begin
  CreateJavaVM;
  Result := FEnvironment^.ExceptionOccurred(FEnvironment);
end;

procedure TJavaBridge.ExceptionDescribe;
begin
  CreateJavaVM;
  FEnvironment^.ExceptionDescribe(FEnvironment);
end;

procedure TJavaBridge.ExceptionClear;
begin
  CreateJavaVM;
  FEnvironment^.ExceptionClear(FEnvironment);
end;

procedure TJavaBridge.FatalError(const Msg: PAnsiChar);
begin
  CreateJavaVM;
  FEnvironment^.FatalError(FEnvironment, Msg);
end;

function TJavaBridge.PushLocalFrame(Capacity: JInt): JInt;
begin
  CreateJavaVM;
  Result := FEnvironment^.PushLocalFrame(FEnvironment, Capacity);
end;

function TJavaBridge.PopLocalFrame(Result_: JObject): JObject;
begin
  CreateJavaVM;
  Result := FEnvironment^.PopLocalFrame(FEnvironment, Result_);
end;

function TJavaBridge.NewGlobalRef(LObj: JObject): JObject;
begin
  CreateJavaVM;
  Result := FEnvironment^.NewGlobalRef(FEnvironment, LObj);
end;

procedure TJavaBridge.DeleteGlobalRef(GRef: JObject);
begin
  CreateJavaVM;
  FEnvironment^.DeleteGlobalRef(FEnvironment, GRef);
end;

procedure TJavaBridge.DeleteLocalRef(Obj: JObject);
begin
  CreateJavaVM;
  FEnvironment^.DeleteLocalRef(FEnvironment, Obj);
end;

function TJavaBridge.IsSameObject(Obj1, Obj2: JObject): JBoolean;
begin
  CreateJavaVM;
  Result := FEnvironment^.IsSameObject(FEnvironment, Obj1, Obj2);
end;

function TJavaBridge.NewLocalRef(Ref: JObject): JObject;
begin
  CreateJavaVM;
  Result := FEnvironment^.NewLocalRef(FEnvironment, Ref);
end;

function TJavaBridge.EnsureLocalCapacity(Capacity: JInt): JInt;
begin
  CreateJavaVM;
  Result := FEnvironment^.EnsureLocalCapacity(FEnvironment, Capacity);
end;

function TJavaBridge.AllocObject(Clazz: JClass): JObject;
begin
  CreateJavaVM;
  Result := FEnvironment^.AllocObject(FEnvironment, Clazz);
end;

function TJavaBridge.NewObject(Clazz: JClass; MethodID: JMethodID { ... } ): JObject;
begin
  CreateJavaVM;
  Result := FEnvironment^.NewObject(FEnvironment, Clazz, MethodID);
end;

function TJavaBridge.NewObjectV(Clazz: JClass; MethodID: JMethodID; Args: VaList): JObject;
begin
  CreateJavaVM;
  Result := FEnvironment^.NewObjectV(FEnvironment, Clazz, MethodID, Args);
end;

function TJavaBridge.NewObjectA(Clazz: JClass; MethodID: JMethodID; const Args: PJValue): JObject;
begin
  CreateJavaVM;
  Result := FEnvironment^.NewObjectA(FEnvironment, Clazz, MethodID, Args);
end;

function TJavaBridge.GetObjectClass(Obj: JObject): JClass;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetObjectClass(FEnvironment, Obj);
end;

function TJavaBridge.IsInstanceOf(Obj: JObject; Clazz: JClass): JBoolean;
begin
  CreateJavaVM;
  Result := FEnvironment^.IsInstanceOf(FEnvironment, Obj, Clazz);
end;

function TJavaBridge.GetMethodID(Clazz: JClass; const Name, Sig: PAnsiChar): JMethodID;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetMethodID(FEnvironment, Clazz, Name, Sig);
end;

function TJavaBridge.CallObjectMethod(Obj: JObject; MethodID: JMethodID { ... } ): JObject;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallObjectMethod(FEnvironment, Obj, MethodID);
end;

function TJavaBridge.CallObjectMethodV(Obj: JObject; MethodID: JMethodID; Args: VaList): JObject;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallObjectMethodV(FEnvironment, Obj, MethodID, Args);
end;

function TJavaBridge.CallObjectMethodA(Obj: JObject; MethodID: JMethodID; const Args: PJValue): JObject;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallObjectMethodA(FEnvironment, Obj, MethodID, Args);
end;

function TJavaBridge.CallBooleanMethod(Obj: JObject; MethodID: JMethodID { ... } ): JBoolean;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallBooleanMethod(FEnvironment, Obj, MethodID);
end;

function TJavaBridge.CallBooleanMethodV(Obj: JObject; MethodID: JMethodID; Args: VaList): JBoolean;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallBooleanMethodV(FEnvironment, Obj, MethodID, Args);
end;

function TJavaBridge.CallBooleanMethodA(Obj: JObject; MethodID: JMethodID; const Args: PJValue): JBoolean;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallBooleanMethodA(FEnvironment, Obj, MethodID, Args);
end;

function TJavaBridge.CallByteMethod(Obj: JObject; MethodID: JMethodID { ... } ): JByte;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallByteMethod(FEnvironment, Obj, MethodID);
end;

function TJavaBridge.CallByteMethodV(Obj: JObject; MethodID: JMethodID; Args: VaList): JByte;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallByteMethodV(FEnvironment, Obj, MethodID, Args);
end;

function TJavaBridge.CallByteMethodA(Obj: JObject; MethodID: JMethodID; const Args: PJValue): JByte;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallByteMethodA(FEnvironment, Obj, MethodID, Args);
end;

function TJavaBridge.CallCharMethod(Obj: JObject; MethodID: JMethodID { ... } ): JChar;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallCharMethod(FEnvironment, Obj, MethodID);
end;

function TJavaBridge.CallCharMethodV(Obj: JObject; MethodID: JMethodID; Args: VaList): JChar;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallCharMethodV(FEnvironment, Obj, MethodID, Args);
end;

function TJavaBridge.CallCharMethodA(Obj: JObject; MethodID: JMethodID; const Args: PJValue): JChar;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallCharMethodA(FEnvironment, Obj, MethodID, Args);
end;

function TJavaBridge.CallShortMethod(Obj: JObject; MethodID: JMethodID { ... } ): JShort;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallShortMethod(FEnvironment, Obj, MethodID);
end;

function TJavaBridge.CallShortMethodV(Obj: JObject; MethodID: JMethodID; Args: VaList): JShort;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallShortMethodV(FEnvironment, Obj, MethodID, Args);
end;

function TJavaBridge.CallShortMethodA(Obj: JObject; MethodID: JMethodID; const Args: PJValue): JShort;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallShortMethodA(FEnvironment, Obj, MethodID, Args);
end;

function TJavaBridge.CallIntMethod(Obj: JObject; MethodID: JMethodID { ... } ): JInt;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallIntMethod(FEnvironment, Obj, MethodID);
end;

function TJavaBridge.CallIntMethodV(Obj: JObject; MethodID: JMethodID; Args: VaList): JInt;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallIntMethodV(FEnvironment, Obj, MethodID, Args);
end;

function TJavaBridge.CallIntMethodA(Obj: JObject; MethodID: JMethodID; const Args: PJValue): JInt;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallIntMethodA(FEnvironment, Obj, MethodID, Args);
end;

function TJavaBridge.CallLongMethod(Obj: JObject; MethodID: JMethodID { ... } ): JLong;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallLongMethod(FEnvironment, Obj, MethodID);
end;

function TJavaBridge.CallLongMethodV(Obj: JObject; MethodID: JMethodID; Args: VaList): JLong;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallLongMethodV(FEnvironment, Obj, MethodID, Args);
end;

function TJavaBridge.CallLongMethodA(Obj: JObject; MethodID: JMethodID; const Args: PJValue): JLong;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallLongMethodA(FEnvironment, Obj, MethodID, Args);
end;

function TJavaBridge.CallFloatMethod(Obj: JObject; MethodID: JMethodID { ... } ): JFloat;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallFloatMethod(FEnvironment, Obj, MethodID);
end;

function TJavaBridge.CallFloatMethodV(Obj: JObject; MethodID: JMethodID; Args: VaList): JFloat;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallFloatMethodV(FEnvironment, Obj, MethodID, Args);
end;

function TJavaBridge.CallFloatMethodA(Obj: JObject; MethodID: JMethodID; const Args: PJValue): JFloat;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallFloatMethodA(FEnvironment, Obj, MethodID, Args);
end;

function TJavaBridge.CallDoubleMethod(Obj: JObject; MethodID: JMethodID { ... } ): JDouble;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallDoubleMethod(FEnvironment, Obj, MethodID);
end;

function TJavaBridge.CallDoubleMethodV(Obj: JObject; MethodID: JMethodID; Args: VaList): JDouble;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallDoubleMethodV(FEnvironment, Obj, MethodID, Args);
end;

function TJavaBridge.CallDoubleMethodA(Obj: JObject; MethodID: JMethodID; const Args: PJValue): JDouble;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallDoubleMethodA(FEnvironment, Obj, MethodID, Args);
end;

procedure TJavaBridge.CallVoidMethod(Obj: JObject; MethodID: JMethodID { ... } );
begin
  CreateJavaVM;
  FEnvironment^.CallVoidMethod(FEnvironment, Obj, MethodID);
end;

procedure TJavaBridge.CallVoidMethodV(Obj: JObject; MethodID: JMethodID; Args: VaList);
begin
  CreateJavaVM;
  FEnvironment^.CallVoidMethodV(FEnvironment, Obj, MethodID, Args);
end;

procedure TJavaBridge.CallVoidMethodA(Obj: JObject; MethodID: JMethodID; const Args: PJValue);
begin
  CreateJavaVM;
  FEnvironment^.CallVoidMethodA(FEnvironment, Obj, MethodID, Args);
end;

function TJavaBridge.CallNonvirtualObjectMethod(Obj: JObject; Clazz: JClass; MethodID: JMethodID { ... } ): JObject;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallNonvirtualObjectMethod(FEnvironment, Obj, Clazz, MethodID);
end;

function TJavaBridge.CallNonvirtualObjectMethodV(Obj: JObject; Clazz: JClass; MethodID: JMethodID; Args: VaList): JObject;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallNonvirtualObjectMethodV(FEnvironment, Obj, Clazz, MethodID, Args);
end;

function TJavaBridge.CallNonvirtualObjectMethodA(Obj: JObject; Clazz: JClass; MethodID: JMethodID; const Args: PJValue): JObject;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallNonvirtualObjectMethodA(FEnvironment, Obj, Clazz, MethodID, Args);
end;

function TJavaBridge.CallNonvirtualBooleanMethod(Obj: JObject; Clazz: JClass; MethodID: JMethodID { ... } ): JBoolean;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallNonvirtualBooleanMethod(FEnvironment, Obj, Clazz, MethodID);
end;

function TJavaBridge.CallNonvirtualBooleanMethodV(Obj: JObject; Clazz: JClass; MethodID: JMethodID; Args: VaList): JBoolean;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallNonvirtualBooleanMethodV(FEnvironment, Obj, Clazz, MethodID, Args);
end;

function TJavaBridge.CallNonvirtualBooleanMethodA(Obj: JObject; Clazz: JClass; MethodID: JMethodID; const Args: PJValue): JBoolean;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallNonvirtualBooleanMethodA(FEnvironment, Obj, Clazz, MethodID, Args);
end;

function TJavaBridge.CallNonvirtualByteMethod(Obj: JObject; Clazz: JClass; MethodID: JMethodID { ... } ): JByte;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallNonvirtualByteMethod(FEnvironment, Obj, Clazz, MethodID);
end;

function TJavaBridge.CallNonvirtualByteMethodV(Obj: JObject; Clazz: JClass; MethodID: JMethodID; Args: VaList): JByte;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallNonvirtualByteMethodV(FEnvironment, Obj, Clazz, MethodID, Args);
end;

function TJavaBridge.CallNonvirtualByteMethodA(Obj: JObject; Clazz: JClass; MethodID: JMethodID; const Args: PJValue): JByte;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallNonvirtualByteMethodA(FEnvironment, Obj, Clazz, MethodID, Args);
end;

function TJavaBridge.CallNonvirtualCharMethod(Obj: JObject; Clazz: JClass; MethodID: JMethodID { ... } ): JChar;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallNonvirtualCharMethod(FEnvironment, Obj, Clazz, MethodID);
end;

function TJavaBridge.CallNonvirtualCharMethodV(Obj: JObject; Clazz: JClass; MethodID: JMethodID; Args: VaList): JChar;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallNonvirtualCharMethodV(FEnvironment, Obj, Clazz, MethodID, Args);
end;

function TJavaBridge.CallNonvirtualCharMethodA(Obj: JObject; Clazz: JClass; MethodID: JMethodID; const Args: PJValue): JChar;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallNonvirtualCharMethodA(FEnvironment, Obj, Clazz, MethodID, Args);
end;

function TJavaBridge.CallNonvirtualShortMethod(Obj: JObject; Clazz: JClass; MethodID: JMethodID { ... } ): JShort;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallNonvirtualShortMethod(FEnvironment, Obj, Clazz, MethodID);
end;

function TJavaBridge.CallNonvirtualShortMethodV(Obj: JObject; Clazz: JClass; MethodID: JMethodID; Args: VaList): JShort;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallNonvirtualShortMethodV(FEnvironment, Obj, Clazz, MethodID, Args);
end;

function TJavaBridge.CallNonvirtualShortMethodA(Obj: JObject; Clazz: JClass; MethodID: JMethodID; const Args: PJValue): JShort;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallNonvirtualShortMethodA(FEnvironment, Obj, Clazz, MethodID, Args);
end;

function TJavaBridge.CallNonvirtualIntMethod(Obj: JObject; Clazz: JClass; MethodID: JMethodID { ... } ): JInt;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallNonvirtualIntMethod(FEnvironment, Obj, Clazz, MethodID);
end;

function TJavaBridge.CallNonvirtualIntMethodV(Obj: JObject; Clazz: JClass; MethodID: JMethodID; Args: VaList): JInt;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallNonvirtualIntMethodV(FEnvironment, Obj, Clazz, MethodID, Args);
end;

function TJavaBridge.CallNonvirtualIntMethodA(Obj: JObject; Clazz: JClass; MethodID: JMethodID; const Args: PJValue): JInt;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallNonvirtualIntMethodA(FEnvironment, Obj, Clazz, MethodID, Args);
end;

function TJavaBridge.CallNonvirtualLongMethod(Obj: JObject; Clazz: JClass; MethodID: JMethodID { ... } ): JLong;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallNonvirtualLongMethod(FEnvironment, Obj, Clazz, MethodID);
end;

function TJavaBridge.CallNonvirtualLongMethodV(Obj: JObject; Clazz: JClass; MethodID: JMethodID; Args: VaList): JLong;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallNonvirtualLongMethodV(FEnvironment, Obj, Clazz, MethodID, Args);
end;

function TJavaBridge.CallNonvirtualLongMethodA(Obj: JObject; Clazz: JClass; MethodID: JMethodID; const Args: PJValue): JLong;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallNonvirtualLongMethodA(FEnvironment, Obj, Clazz, MethodID, Args);
end;

function TJavaBridge.CallNonvirtualFloatMethod(Obj: JObject; Clazz: JClass; MethodID: JMethodID { ... } ): JFloat;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallNonvirtualFloatMethod(FEnvironment, Obj, Clazz, MethodID);
end;

function TJavaBridge.CallNonvirtualFloatMethodV(Obj: JObject; Clazz: JClass; MethodID: JMethodID; Args: VaList): JFloat;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallNonvirtualFloatMethodV(FEnvironment, Obj, Clazz, MethodID, Args);
end;

function TJavaBridge.CallNonvirtualFloatMethodA(Obj: JObject; Clazz: JClass; MethodID: JMethodID; const Args: PJValue): JFloat;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallNonvirtualFloatMethodA(FEnvironment, Obj, Clazz, MethodID, Args);
end;

function TJavaBridge.CallNonvirtualDoubleMethod(Obj: JObject; Clazz: JClass; MethodID: JMethodID { ... } ): JDouble;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallNonvirtualDoubleMethod(FEnvironment, Obj, Clazz, MethodID);
end;

function TJavaBridge.CallNonvirtualDoubleMethodV(Obj: JObject; Clazz: JClass; MethodID: JMethodID; Args: VaList): JDouble;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallNonvirtualDoubleMethodV(FEnvironment, Obj, Clazz, MethodID, Args);
end;

function TJavaBridge.CallNonvirtualDoubleMethodA(Obj: JObject; Clazz: JClass; MethodID: JMethodID; const Args: PJValue): JDouble;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallNonvirtualDoubleMethodA(FEnvironment, Obj, Clazz, MethodID, Args);
end;

procedure TJavaBridge.CallNonvirtualVoidMethod(Obj: JObject; Clazz: JClass; MethodID: JMethodID { ... } );
begin
  CreateJavaVM;
  FEnvironment^.CallNonvirtualVoidMethod(FEnvironment, Obj, Clazz, MethodID);
end;

procedure TJavaBridge.CallNonvirtualVoidMethodV(Obj: JObject; Clazz: JClass; MethodID: JMethodID; Args: VaList);
begin
  CreateJavaVM;
  FEnvironment^.CallNonvirtualVoidMethodV(FEnvironment, Obj, Clazz, MethodID, Args);
end;

procedure TJavaBridge.CallNonvirtualVoidMethodA(Obj: JObject; Clazz: JClass; MethodID: JMethodID; const Args: PJValue);
begin
  CreateJavaVM;
  FEnvironment^.CallNonvirtualVoidMethodA(FEnvironment, Obj, Clazz, MethodID, Args);
end;

function TJavaBridge.GetFieldID(Clazz: JClass; const Name, Sig: PAnsiChar): JFieldID;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetFieldID(FEnvironment, Clazz, Name, Sig);
end;

function TJavaBridge.GetObjectField(Obj: JObject; FieldID: JFieldID): JObject;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetObjectField(FEnvironment, Obj, FieldID);
end;

function TJavaBridge.GetBooleanField(Obj: JObject; FieldID: JFieldID): JBoolean;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetBooleanField(FEnvironment, Obj, FieldID);
end;

function TJavaBridge.GetByteField(Obj: JObject; FieldID: JFieldID): JByte;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetByteField(FEnvironment, Obj, FieldID);
end;

function TJavaBridge.GetCharField(Obj: JObject; FieldID: JFieldID): JChar;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetCharField(FEnvironment, Obj, FieldID);
end;

function TJavaBridge.GetShortField(Obj: JObject; FieldID: JFieldID): JShort;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetShortField(FEnvironment, Obj, FieldID);
end;

function TJavaBridge.GetIntField(Obj: JObject; FieldID: JFieldID): JInt;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetIntField(FEnvironment, Obj, FieldID);
end;

function TJavaBridge.GetLongField(Obj: JObject; FieldID: JFieldID): JLong;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetLongField(FEnvironment, Obj, FieldID);
end;

function TJavaBridge.GetFloatField(Obj: JObject; FieldID: JFieldID): JFloat;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetFloatField(FEnvironment, Obj, FieldID);
end;

function TJavaBridge.GetDoubleField(Obj: JObject; FieldID: JFieldID): JDouble;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetDoubleField(FEnvironment, Obj, FieldID);
end;

procedure TJavaBridge.SetObjectField(Obj: JObject; FieldID: JFieldID; Val: JObject);
begin
  CreateJavaVM;
  FEnvironment^.SetObjectField(FEnvironment, Obj, FieldID, Val);
end;

procedure TJavaBridge.SetBooleanField(Obj: JObject; FieldID: JFieldID; Val: JBoolean);
begin
  CreateJavaVM;
  FEnvironment^.SetBooleanField(FEnvironment, Obj, FieldID, Val);
end;

procedure TJavaBridge.SetByteField(Obj: JObject; FieldID: JFieldID; Val: JByte);
begin
  CreateJavaVM;
  FEnvironment^.SetByteField(FEnvironment, Obj, FieldID, Val);
end;

procedure TJavaBridge.SetCharField(Obj: JObject; FieldID: JFieldID; Val: JChar);
begin
  CreateJavaVM;
  FEnvironment^.SetCharField(FEnvironment, Obj, FieldID, Val);
end;

procedure TJavaBridge.SetShortField(Obj: JObject; FieldID: JFieldID; Val: JShort);
begin
  CreateJavaVM;
  FEnvironment^.SetShortField(FEnvironment, Obj, FieldID, Val);
end;

procedure TJavaBridge.SetIntField(Obj: JObject; FieldID: JFieldID; Val: JInt);
begin
  CreateJavaVM;
  FEnvironment^.SetIntField(FEnvironment, Obj, FieldID, Val);
end;

procedure TJavaBridge.SetLongField(Obj: JObject; FieldID: JFieldID; Val: JLong);
begin
  CreateJavaVM;
  FEnvironment^.SetLongField(FEnvironment, Obj, FieldID, Val);
end;

procedure TJavaBridge.SetFloatField(Obj: JObject; FieldID: JFieldID; Val: JFloat);
begin
  CreateJavaVM;
  FEnvironment^.SetFloatField(FEnvironment, Obj, FieldID, Val);
end;

procedure TJavaBridge.SetDoubleField(Obj: JObject; FieldID: JFieldID; Val: JDouble);
begin
  CreateJavaVM;
  FEnvironment^.SetDoubleField(FEnvironment, Obj, FieldID, Val);
end;

function TJavaBridge.GetStaticMethodID(Clazz: JClass; const Name, Sig: PAnsiChar): JMethodID;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetStaticMethodID(FEnvironment, Clazz, Name, Sig);
end;

function TJavaBridge.CallStaticObjectMethod(Clazz: JClass; MethodID: JMethodID { ... } ): JObject;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallStaticObjectMethod(FEnvironment, Clazz, MethodID);
end;

function TJavaBridge.CallStaticObjectMethodV(Clazz: JClass; MethodID: JMethodID; Args: VaList): JObject;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallStaticObjectMethodV(FEnvironment, Clazz, MethodID, Args);
end;

function TJavaBridge.CallStaticObjectMethodA(Clazz: JClass; MethodID: JMethodID; const Args: PJValue): JObject;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallStaticObjectMethodA(FEnvironment, Clazz, MethodID, Args);
end;

function TJavaBridge.CallStaticBooleanMethod(Clazz: JClass; MethodID: JMethodID { ... } ): JBoolean;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallStaticBooleanMethod(FEnvironment, Clazz, MethodID);
end;

function TJavaBridge.CallStaticBooleanMethodV(Clazz: JClass; MethodID: JMethodID; Args: VaList): JBoolean;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallStaticBooleanMethodV(FEnvironment, Clazz, MethodID, Args);
end;

function TJavaBridge.CallStaticBooleanMethodA(Clazz: JClass; MethodID: JMethodID; const Args: PJValue): JBoolean;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallStaticBooleanMethodA(FEnvironment, Clazz, MethodID, Args);
end;

function TJavaBridge.CallStaticByteMethod(Clazz: JClass; MethodID: JMethodID { ... } ): JByte;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallStaticByteMethod(FEnvironment, Clazz, MethodID);
end;

function TJavaBridge.CallStaticByteMethodV(Clazz: JClass; MethodID: JMethodID; Args: VaList): JByte;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallStaticByteMethodV(FEnvironment, Clazz, MethodID, Args);
end;

function TJavaBridge.CallStaticByteMethodA(Clazz: JClass; MethodID: JMethodID; const Args: PJValue): JByte;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallStaticByteMethodA(FEnvironment, Clazz, MethodID, Args);
end;

function TJavaBridge.CallStaticCharMethod(Clazz: JClass; MethodID: JMethodID { ... } ): JChar;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallStaticCharMethod(FEnvironment, Clazz, MethodID);
end;

function TJavaBridge.CallStaticCharMethodV(Clazz: JClass; MethodID: JMethodID; Args: VaList): JChar;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallStaticCharMethodV(FEnvironment, Clazz, MethodID, Args);
end;

function TJavaBridge.CallStaticCharMethodA(Clazz: JClass; MethodID: JMethodID; const Args: PJValue): JChar;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallStaticCharMethodA(FEnvironment, Clazz, MethodID, Args);
end;

function TJavaBridge.CallStaticShortMethod(Clazz: JClass; MethodID: JMethodID { ... } ): JShort;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallStaticShortMethod(FEnvironment, Clazz, MethodID);
end;

function TJavaBridge.CallStaticShortMethodV(Clazz: JClass; MethodID: JMethodID; Args: VaList): JShort;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallStaticShortMethodV(FEnvironment, Clazz, MethodID, Args);
end;

function TJavaBridge.CallStaticShortMethodA(Clazz: JClass; MethodID: JMethodID; const Args: PJValue): JShort;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallStaticShortMethodA(FEnvironment, Clazz, MethodID, Args);
end;

function TJavaBridge.CallStaticIntMethod(Clazz: JClass; MethodID: JMethodID { ... } ): JInt;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallStaticIntMethod(FEnvironment, Clazz, MethodID);
end;

function TJavaBridge.CallStaticIntMethodV(Clazz: JClass; MethodID: JMethodID; Args: VaList): JInt;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallStaticIntMethodV(FEnvironment, Clazz, MethodID, Args);
end;

function TJavaBridge.CallStaticIntMethodA(Clazz: JClass; MethodID: JMethodID; const Args: PJValue): JInt;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallStaticIntMethodA(FEnvironment, Clazz, MethodID, Args);
end;

function TJavaBridge.CallStaticLongMethod(Clazz: JClass; MethodID: JMethodID { ... } ): JLong;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallStaticLongMethod(FEnvironment, Clazz, MethodID);
end;

function TJavaBridge.CallStaticLongMethodV(Clazz: JClass; MethodID: JMethodID; Args: VaList): JLong;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallStaticLongMethodV(FEnvironment, Clazz, MethodID, Args);
end;

function TJavaBridge.CallStaticLongMethodA(Clazz: JClass; MethodID: JMethodID; const Args: PJValue): JLong;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallStaticLongMethodA(FEnvironment, Clazz, MethodID, Args);
end;

function TJavaBridge.CallStaticFloatMethod(Clazz: JClass; MethodID: JMethodID { ... } ): JFloat;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallStaticFloatMethod(FEnvironment, Clazz, MethodID);
end;

function TJavaBridge.CallStaticFloatMethodV(Clazz: JClass; MethodID: JMethodID; Args: VaList): JFloat;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallStaticFloatMethodV(FEnvironment, Clazz, MethodID, Args);
end;

function TJavaBridge.CallStaticFloatMethodA(Clazz: JClass; MethodID: JMethodID; const Args: PJValue): JFloat;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallStaticFloatMethodA(FEnvironment, Clazz, MethodID, Args);
end;

function TJavaBridge.CallStaticDoubleMethod(Clazz: JClass; MethodID: JMethodID { ... } ): JDouble;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallStaticDoubleMethod(FEnvironment, Clazz, MethodID);
end;

function TJavaBridge.CallStaticDoubleMethodV(Clazz: JClass; MethodID: JMethodID; Args: VaList): JDouble;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallStaticDoubleMethodV(FEnvironment, Clazz, MethodID, Args);
end;

function TJavaBridge.CallStaticDoubleMethodA(Clazz: JClass; MethodID: JMethodID; const Args: PJValue): JDouble;
begin
  CreateJavaVM;
  Result := FEnvironment^.CallStaticDoubleMethodA(FEnvironment, Clazz, MethodID, Args);
end;

procedure TJavaBridge.CallStaticVoidMethod(Cls: JClass; MethodID: JMethodID { ... } );
begin
  CreateJavaVM;
  FEnvironment^.CallStaticVoidMethod(FEnvironment, Cls, MethodID);
end;

procedure TJavaBridge.CallStaticVoidMethodV(Cls: JClass; MethodID: JMethodID; Args: VaList);
begin
  CreateJavaVM;
  FEnvironment^.CallStaticVoidMethodV(FEnvironment, Cls, MethodID, Args);
end;

procedure TJavaBridge.CallStaticVoidMethodA(Cls: JClass; MethodID: JMethodID; const Args: PJValue);
begin
  CreateJavaVM;
  FEnvironment^.CallStaticVoidMethodA(FEnvironment, Cls, MethodID, Args);
end;

function TJavaBridge.GetStaticFieldID(Clazz: JClass; const Name, Sig: PAnsiChar): JFieldID;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetStaticFieldID(FEnvironment, Clazz, Name, Sig);
end;

function TJavaBridge.GetStaticObjectField(Clazz: JClass; FieldID: JFieldID): JObject;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetStaticObjectField(FEnvironment, Clazz, FieldID);
end;

function TJavaBridge.GetStaticBooleanField(Clazz: JClass; FieldID: JFieldID): JBoolean;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetStaticBooleanField(FEnvironment, Clazz, FieldID);
end;

function TJavaBridge.GetStaticByteField(Clazz: JClass; FieldID: JFieldID): JByte;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetStaticByteField(FEnvironment, Clazz, FieldID);
end;

function TJavaBridge.GetStaticCharField(Clazz: JClass; FieldID: JFieldID): JChar;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetStaticCharField(FEnvironment, Clazz, FieldID);
end;

function TJavaBridge.GetStaticShortField(Clazz: JClass; FieldID: JFieldID): JShort;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetStaticShortField(FEnvironment, Clazz, FieldID);
end;

function TJavaBridge.GetStaticIntField(Clazz: JClass; FieldID: JFieldID): JInt;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetStaticIntField(FEnvironment, Clazz, FieldID);
end;

function TJavaBridge.GetStaticLongField(Clazz: JClass; FieldID: JFieldID): JLong;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetStaticLongField(FEnvironment, Clazz, FieldID);
end;

function TJavaBridge.GetStaticFloatField(Clazz: JClass; FieldID: JFieldID): JFloat;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetStaticFloatField(FEnvironment, Clazz, FieldID);
end;

function TJavaBridge.GetStaticDoubleField(Clazz: JClass; FieldID: JFieldID): JDouble;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetStaticDoubleField(FEnvironment, Clazz, FieldID);
end;

procedure TJavaBridge.SetStaticObjectField(Clazz: JClass; FieldID: JFieldID; Value: JObject);
begin
  CreateJavaVM;
  FEnvironment^.SetStaticObjectField(FEnvironment, Clazz, FieldID, Value);
end;

procedure TJavaBridge.SetStaticBooleanField(Clazz: JClass; FieldID: JFieldID; Value: JBoolean);
begin
  CreateJavaVM;
  FEnvironment^.SetStaticBooleanField(FEnvironment, Clazz, FieldID, Value);
end;

procedure TJavaBridge.SetStaticByteField(Clazz: JClass; FieldID: JFieldID; Value: JByte);
begin
  CreateJavaVM;
  FEnvironment^.SetStaticByteField(FEnvironment, Clazz, FieldID, Value);
end;

procedure TJavaBridge.SetStaticCharField(Clazz: JClass; FieldID: JFieldID; Value: JChar);
begin
  CreateJavaVM;
  FEnvironment^.SetStaticCharField(FEnvironment, Clazz, FieldID, Value);
end;

procedure TJavaBridge.SetStaticShortField(Clazz: JClass; FieldID: JFieldID; Value: JShort);
begin
  CreateJavaVM;
  FEnvironment^.SetStaticShortField(FEnvironment, Clazz, FieldID, Value);
end;

procedure TJavaBridge.SetStaticIntField(Clazz: JClass; FieldID: JFieldID; Value: JInt);
begin
  CreateJavaVM;
  FEnvironment^.SetStaticIntField(FEnvironment, Clazz, FieldID, Value);
end;

procedure TJavaBridge.SetStaticLongField(Clazz: JClass; FieldID: JFieldID; Value: JLong);
begin
  CreateJavaVM;
  FEnvironment^.SetStaticLongField(FEnvironment, Clazz, FieldID, Value);
end;

procedure TJavaBridge.SetStaticFloatField(Clazz: JClass; FieldID: JFieldID; Value: JFloat);
begin
  CreateJavaVM;
  FEnvironment^.SetStaticFloatField(FEnvironment, Clazz, FieldID, Value);
end;

procedure TJavaBridge.SetStaticDoubleField(Clazz: JClass; FieldID: JFieldID; Value: JDouble);
begin
  CreateJavaVM;
  FEnvironment^.SetStaticDoubleField(FEnvironment, Clazz, FieldID, Value);
end;

function TJavaBridge.NewString(const Unicode: PJChar; Len: JSize): JString;
begin
  CreateJavaVM;
  Result := FEnvironment^.NewString(FEnvironment, Unicode, Len);
end;

function TJavaBridge.GetStringLength(Str: JString): JSize;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetStringLength(FEnvironment, Str);
end;

function TJavaBridge.GetStringChars(Str: JString; var IsCopy: JBoolean): PJChar;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetStringChars(FEnvironment, Str, IsCopy);
end;

procedure TJavaBridge.ReleaseStringChars(Str: JString; const Chars: PJChar);
begin
  CreateJavaVM;
  FEnvironment^.ReleaseStringChars(FEnvironment, Str, Chars);
end;

function TJavaBridge.NewStringUTF(const Utf: PAnsiChar): JString;
begin
  CreateJavaVM;
  Result := FEnvironment^.NewStringUTF(FEnvironment, Utf);
end;

function TJavaBridge.GetStringUTFLength(Str: JString): JSize;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetStringUTFLength(FEnvironment, Str);
end;

function TJavaBridge.GetStringUTFChars(Str: JString; var IsCopy: JBoolean): PAnsiChar;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetStringUTFChars(FEnvironment, Str, IsCopy);
end;

procedure TJavaBridge.ReleaseStringUTFChars(Str: JString; const Chars: PAnsiChar);
begin
  CreateJavaVM;
  FEnvironment^.ReleaseStringUTFChars(FEnvironment, Str, Chars);
end;

function TJavaBridge.GetArrayLength(Array_: JArray): JSize;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetArrayLength(FEnvironment, Array_);
end;

function TJavaBridge.NewObjectArray(Len: JSize; Clazz: JClass; Init: JObject): JObjectArray;
begin
  CreateJavaVM;
  Result := FEnvironment^.NewObjectArray(FEnvironment, Len, Clazz, Init);
end;

function TJavaBridge.GetObjectArrayElement(Array_: JObjectArray; Index: JSize): JObject;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetObjectArrayElement(FEnvironment, Array_, Index);
end;

procedure TJavaBridge.SetObjectArrayElement(Array_: JObjectArray; Index: JSize; Val: JObject);
begin
  CreateJavaVM;
  FEnvironment^.SetObjectArrayElement(FEnvironment, Array_, Index, Val);
end;

function TJavaBridge.NewBooleanArray(Len: JSize): JBooleanArray;
begin
  CreateJavaVM;
  Result := FEnvironment^.NewBooleanArray(FEnvironment, Len);
end;

function TJavaBridge.NewByteArray(Len: JSize): JByteArray;
begin
  CreateJavaVM;
  Result := FEnvironment^.NewByteArray(FEnvironment, Len);
end;

function TJavaBridge.NewCharArray(Len: JSize): JCharArray;
begin
  CreateJavaVM;
  Result := FEnvironment^.NewCharArray(FEnvironment, Len);
end;

function TJavaBridge.NewShortArray(Len: JSize): JShortArray;
begin
  CreateJavaVM;
  Result := FEnvironment^.NewShortArray(FEnvironment, Len);
end;

function TJavaBridge.NewIntArray(Len: JSize): JIntArray;
begin
  CreateJavaVM;
  Result := FEnvironment^.NewIntArray(FEnvironment, Len);
end;

function TJavaBridge.NewLongArray(Len: JSize): JLongArray;
begin
  CreateJavaVM;
  Result := FEnvironment^.NewLongArray(FEnvironment, Len);
end;

function TJavaBridge.NewFloatArray(Len: JSize): JFloatArray;
begin
  CreateJavaVM;
  Result := FEnvironment^.NewFloatArray(FEnvironment, Len);
end;

function TJavaBridge.NewDoubleArray(Len: JSize): JDoubleArray;
begin
  CreateJavaVM;
  Result := FEnvironment^.NewDoubleArray(FEnvironment, Len);
end;

function TJavaBridge.GetBooleanArrayElements(Array_: JBooleanArray; var IsCopy: JBoolean): PJBoolean;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetBooleanArrayElements(FEnvironment, Array_, IsCopy);
end;

function TJavaBridge.GetByteArrayElements(Array_: JByteArray; var IsCopy: JBoolean): PJByte;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetByteArrayElements(FEnvironment, Array_, IsCopy);
end;

function TJavaBridge.GetCharArrayElements(Array_: JCharArray; var IsCopy: JBoolean): PJChar;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetCharArrayElements(FEnvironment, Array_, IsCopy);
end;

function TJavaBridge.GetShortArrayElements(Array_: JShortArray; var IsCopy: JBoolean): PJShort;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetShortArrayElements(FEnvironment, Array_, IsCopy);
end;

function TJavaBridge.GetIntArrayElements(Array_: JIntArray; var IsCopy: JBoolean): PJInt;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetIntArrayElements(FEnvironment, Array_, IsCopy);
end;

function TJavaBridge.GetLongArrayElements(Array_: JLongArray; var IsCopy: JBoolean): PJLong;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetLongArrayElements(FEnvironment, Array_, IsCopy);
end;

function TJavaBridge.GetFloatArrayElements(Array_: JFloatArray; var IsCopy: JBoolean): PJFloat;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetFloatArrayElements(FEnvironment, Array_, IsCopy);
end;

function TJavaBridge.GetDoubleArrayElements(Array_: JDoubleArray; var IsCopy: JBoolean): PJDouble;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetDoubleArrayElements(FEnvironment, Array_, IsCopy);
end;

procedure TJavaBridge.ReleaseBooleanArrayElements(Array_: JBooleanArray; Elems: PJBoolean; Mode: JInt);
begin
  CreateJavaVM;
  FEnvironment^.ReleaseBooleanArrayElements(FEnvironment, Array_, Elems, Mode);
end;

procedure TJavaBridge.ReleaseByteArrayElements(Array_: JByteArray; Elems: PJByte; Mode: JInt);
begin
  CreateJavaVM;
  FEnvironment^.ReleaseByteArrayElements(FEnvironment, Array_, Elems, Mode);
end;

procedure TJavaBridge.ReleaseCharArrayElements(Array_: JCharArray; Elems: PJChar; Mode: JInt);
begin
  CreateJavaVM;
  FEnvironment^.ReleaseCharArrayElements(FEnvironment, Array_, Elems, Mode);
end;

procedure TJavaBridge.ReleaseShortArrayElements(Array_: JShortArray; Elems: PJShort; Mode: JInt);
begin
  CreateJavaVM;
  FEnvironment^.ReleaseShortArrayElements(FEnvironment, Array_, Elems, Mode);
end;

procedure TJavaBridge.ReleaseIntArrayElements(Array_: JIntArray; Elems: PJInt; Mode: JInt);
begin
  CreateJavaVM;
  FEnvironment^.ReleaseIntArrayElements(FEnvironment, Array_, Elems, Mode);
end;

procedure TJavaBridge.ReleaseLongArrayElements(Array_: JLongArray; Elems: PJLong; Mode: JInt);
begin
  CreateJavaVM;
  FEnvironment^.ReleaseLongArrayElements(FEnvironment, Array_, Elems, Mode);
end;

procedure TJavaBridge.ReleaseFloatArrayElements(Array_: JFloatArray; Elems: PJFloat; Mode: JInt);
begin
  CreateJavaVM;
  FEnvironment^.ReleaseFloatArrayElements(FEnvironment, Array_, Elems, Mode);
end;

procedure TJavaBridge.ReleaseDoubleArrayElements(Array_: JDoubleArray; Elems: PJDouble; Mode: JInt);
begin
  CreateJavaVM;
  FEnvironment^.ReleaseDoubleArrayElements(FEnvironment, Array_, Elems, Mode);
end;

procedure TJavaBridge.GetBooleanArrayRegion(Array_: JBooleanArray; Start, Len: JSize; Buf: PJBoolean);
begin
  CreateJavaVM;
  FEnvironment^.GetBooleanArrayRegion(FEnvironment, Array_, Start, Len, Buf);
end;

procedure TJavaBridge.GetByteArrayRegion(Array_: JByteArray; Start, Len: JSize; Buf: PJByte);
begin
  CreateJavaVM;
  FEnvironment^.GetByteArrayRegion(FEnvironment, Array_, Start, Len, Buf);
end;

procedure TJavaBridge.GetCharArrayRegion(Array_: JCharArray; Start, Len: JSize; Buf: PJChar);
begin
  CreateJavaVM;
  FEnvironment^.GetCharArrayRegion(FEnvironment, Array_, Start, Len, Buf);
end;

procedure TJavaBridge.GetShortArrayRegion(Array_: JShortArray; Start, Len: JSize; Buf: PJShort);
begin
  CreateJavaVM;
  FEnvironment^.GetShortArrayRegion(FEnvironment, Array_, Start, Len, Buf);
end;

procedure TJavaBridge.GetIntArrayRegion(Array_: JIntArray; Start, Len: JSize; Buf: PJInt);
begin
  CreateJavaVM;
  FEnvironment^.GetIntArrayRegion(FEnvironment, Array_, Start, Len, Buf);
end;

procedure TJavaBridge.GetLongArrayRegion(Array_: JLongArray; Start, Len: JSize; Buf: PJLong);
begin
  CreateJavaVM;
  FEnvironment^.GetLongArrayRegion(FEnvironment, Array_, Start, Len, Buf);
end;

procedure TJavaBridge.GetFloatArrayRegion(Array_: JFloatArray; Start, Len: JSize; Buf: PJFloat);
begin
  CreateJavaVM;
  FEnvironment^.GetFloatArrayRegion(FEnvironment, Array_, Start, Len, Buf);
end;

procedure TJavaBridge.GetDoubleArrayRegion(Array_: JDoubleArray; Start, Len: JSize; Buf: PJDouble);
begin
  CreateJavaVM;
  FEnvironment^.GetDoubleArrayRegion(FEnvironment, Array_, Start, Len, Buf);
end;

procedure TJavaBridge.SetBooleanArrayRegion(Array_: JBooleanArray; Start, Len: JSize; const Buf: PJBoolean);
begin
  CreateJavaVM;
  FEnvironment^.SetBooleanArrayRegion(FEnvironment, Array_, Start, Len, Buf);
end;

procedure TJavaBridge.SetByteArrayRegion(Array_: JByteArray; Start, Len: JSize; const Buf: PJByte);
begin
  CreateJavaVM;
  FEnvironment^.SetByteArrayRegion(FEnvironment, Array_, Start, Len, Buf);
end;

procedure TJavaBridge.SetCharArrayRegion(Array_: JCharArray; Start, Len: JSize; const Buf: PJChar);
begin
  CreateJavaVM;
  FEnvironment^.SetCharArrayRegion(FEnvironment, Array_, Start, Len, Buf);
end;

procedure TJavaBridge.SetShortArrayRegion(Array_: JShortArray; Start, Len: JSize; const Buf: PJShort);
begin
  CreateJavaVM;
  FEnvironment^.SetShortArrayRegion(FEnvironment, Array_, Start, Len, Buf);
end;

procedure TJavaBridge.SetIntArrayRegion(Array_: JIntArray; Start, Len: JSize; const Buf: PJInt);
begin
  CreateJavaVM;
  FEnvironment^.SetIntArrayRegion(FEnvironment, Array_, Start, Len, Buf);
end;

procedure TJavaBridge.SetLongArrayRegion(Array_: JLongArray; Start, Len: JSize; const Buf: PJLong);
begin
  CreateJavaVM;
  FEnvironment^.SetLongArrayRegion(FEnvironment, Array_, Start, Len, Buf);
end;

procedure TJavaBridge.SetFloatArrayRegion(Array_: JFloatArray; Start, Len: JSize; const Buf: PJFloat);
begin
  CreateJavaVM;
  FEnvironment^.SetFloatArrayRegion(FEnvironment, Array_, Start, Len, Buf);
end;

procedure TJavaBridge.SetDoubleArrayRegion(Array_: JDoubleArray; Start, Len: JSize; const Buf: PJDouble);
begin
  CreateJavaVM;
  FEnvironment^.SetDoubleArrayRegion(FEnvironment, Array_, Start, Len, Buf);
end;

function TJavaBridge.RegisterNatives(Clazz: JClass; const Methods: PJNINativeMethod; NMethods: JInt): JInt;
begin
  CreateJavaVM;
  Result := FEnvironment^.RegisterNatives(FEnvironment, Clazz, Methods, NMethods);
end;

function TJavaBridge.UnregisterNatives(Clazz: JClass): JInt;
begin
  CreateJavaVM;
  Result := FEnvironment^.UnregisterNatives(FEnvironment, Clazz);
end;

function TJavaBridge.MonitorEnter(Obj: JObject): JInt;
begin
  CreateJavaVM;
  Result := FEnvironment^.MonitorEnter(FEnvironment, Obj);
end;

function TJavaBridge.MonitorExit(Obj: JObject): JInt;
begin
  CreateJavaVM;
  Result := FEnvironment^.MonitorExit(FEnvironment, Obj);
end;

function TJavaBridge.GetJavaVM(out VM: PJavaVM): JInt;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetJavaVM(FEnvironment, VM);
end;

procedure TJavaBridge.GetStringRegion(Str: JString; Start, Len: JSize; Buf: PJChar);
begin
  CreateJavaVM;
  FEnvironment^.GetStringRegion(FEnvironment, Str, Start, Len, Buf);
end;

procedure TJavaBridge.GetStringUTFRegion(Str: JString; Start, Len: JSize; Buf: PAnsiChar);
begin
  CreateJavaVM;
  FEnvironment^.GetStringUTFRegion(FEnvironment, Str, Start, Len, Buf);
end;

function TJavaBridge.GetPrimitiveArrayCritical(Array_: JArray; var IsCopy: JBoolean): Pointer;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetPrimitiveArrayCritical(FEnvironment, Array_, IsCopy);
end;

procedure TJavaBridge.ReleasePrimitiveArrayCritical(Array_: JArray; Carray: Pointer; Mode: JInt);
begin
  CreateJavaVM;
  FEnvironment^.ReleasePrimitiveArrayCritical(FEnvironment, Array_, Carray, Mode);
end;

function TJavaBridge.GetStringCritical(String_: JString; var IsCopy: JBoolean): PJChar;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetStringCritical(FEnvironment, String_, IsCopy);
end;

procedure TJavaBridge.ReleaseStringCritical(String_: JString; const CString: PJChar);
begin
  CreateJavaVM;
  FEnvironment^.ReleaseStringCritical(FEnvironment, String_, CString);
end;

function TJavaBridge.NewWeakGlobalRef(Obj: JObject): JWeak;
begin
  CreateJavaVM;
  Result := FEnvironment^.NewWeakGlobalRef(FEnvironment, Obj);
end;

procedure TJavaBridge.DeleteWeakGlobalRef(Ref: JWeak);
begin
  CreateJavaVM;
  FEnvironment^.DeleteWeakGlobalRef(FEnvironment, Ref);
end;

function TJavaBridge.ExceptionCheck: JBoolean;
begin
  CreateJavaVM;
  Result := FEnvironment^.ExceptionCheck(FEnvironment);
end;

function TJavaBridge.NewDirectByteBuffer(Address: Pointer; Capacity: JLong): JObject;
begin
  CreateJavaVM;
  Result := FEnvironment^.NewDirectByteBuffer(FEnvironment, Address, Capacity);
end;

function TJavaBridge.GetDirectBufferAddress(Buf: JObject): Pointer;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetDirectBufferAddress(FEnvironment, Buf);
end;

function TJavaBridge.GetDirectBufferCapacity(Buf: JObject): JLong;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetDirectBufferCapacity(FEnvironment, Buf);
end;

function TJavaBridge.GetObjectRefType(Obj: JObject): JObjectRefType;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetObjectRefType(FEnvironment, Obj);
end;

function TJavaBridge.GetModule(Clazz: JClass): JObject;
begin
  CreateJavaVM;
  Result := FEnvironment^.GetModule(FEnvironment, Clazz);
end;

end.